# -*- encoding:utf-8 -*-
"""
    统计相关工具模块
"""
import numpy as np
import pandas as pd
import functools, six
import scipy.stats as scs
from collections import namedtuple, Iterable
from sklearn.metrics.pairwise import euclidean_distances, manhattan_distances, cosine_distances
from ultron.kdutils.scaler import scaler_mm
from ultron.ump.core.helper import pd_rolling_mean


# noinspection PyClassHasNoInit
class MomentsTuple(
        namedtuple(
            'MomentsTuple',
            ('count', 'max', 'min', 'mean', 'std', 'skewness', 'kurtosis'))):
    __slots__ = ()

    def __repr__(self):
        return "count:{}\nmax:{}\nmin:{}\nmean:{}\nstd:{}\nskewness:{}\nkurtosis:{}".format(
            self.count, self.max, self.min, self.mean, self.std, self.skewness,
            self.kurtosis)


# noinspection PyUnresolvedReferences
def _distance_matrix(distance_func, df, scale_end=True, to_similar=False):
    """
    非两两distance计算，限制只有一个矩阵的输入，且输入必须为pd.DataFrame or np.array or 多层迭代序列[[],[]]
    :param distance_func: 计算距离的方法
    :param df: pd.DataFrame or np.array or 多层迭代序列[[],[]], 之所以叫df，是因为在内部会统一转换为pd.DataFrame
    :param scale_end: 对结果矩阵进行标准化处理
    :param to_similar: 是否进行后置输出转换similar值
    :return: distance_matrix，pd.DataFrame对象
    """

    if not callable(distance_func):
        raise TypeError('distance_func must callable!!!')

    if isinstance(df, np.ndarray):
        # 把np.ndarray转DataFrame，便统一处理
        df = pd.DataFrame(df)

    if not isinstance(df, pd.DataFrame):
        if all(isinstance(arr_item, Iterable) for arr_item in df):
            # 如果子序列的元素也都是可以迭代的，那么先转np.array，然后再DataFrame
            df = pd.DataFrame(np.array(df))
        else:
            raise TypeError('df must pd.DataFrame object!!!')
    # 做列的distance所以df.T
    distance = distance_func(df.T)

    if scale_end:
        # TODO 这里需要可以设置标准化使用的方法，暂时都使用scaler_mm
        distance = scaler_mm(distance)
        if to_similar:
            # 只有scale_end的情况下to_similar才会生效，否则没有意义
            distance = 1 - distance

    # 将计算结果的distance转换为pd.DataFrame对象，行和列索引都使用df.columns
    distance_df = pd.DataFrame(distance, index=df.columns, columns=df.columns)
    return distance_df


def arr_to_pandas(func):
    """
        函数装饰器：定参数装饰器，非通用，通用转换使用ABuDTUtil中的装饰器
        将被装饰函数中的arr序列转换为pd.DataFrame或者pd.Series
    """

    @functools.wraps(func)
    def wrapper(arr, *arg, **kwargs):

        # TODO Iterable和six.string_types的判断抽出来放在一个模块，做为Iterable的判断来使用
        if not isinstance(arr, Iterable) or isinstance(arr, six.string_types):
            # arr必须是可以迭代的对象
            raise TypeError('arr not isinstance of Iterable')

        if not isinstance(arr, pd.DataFrame) or isinstance(arr, pd.Series):
            if isinstance(
                    arr,
                    np.ndarray) and len(arr.shape) > 1 and arr.shape[1] > 1:
                # np.ndarray > 1列的转换为pd.DataFrame
                arr = pd.DataFrame(arr)
            elif isinstance(arr, dict):
                # 针对dict转换pd.DataFrame
                arr = pd.DataFrame(arr)
            elif all(isinstance(arr_item, Iterable) for arr_item in arr):
                # 如果子序列的元素也都是可以迭代的，那么先转np.array，然后再DataFrame
                arr = pd.DataFrame(np.array(arr))
            else:
                # 否则序列对象转换为pd.Series
                arr = pd.Series(arr)
        return func(arr, *arg, **kwargs)

    return wrapper


def arr_to_numpy(func):
    """
        函数装饰器：定参数装饰器，非通用，通用转换使用UltronDTUtil中的装饰器
        将被装饰函数中的arr序列转换为np.array
    """

    @functools.wraps(func)
    def wrapper(arr, *arg, **kwargs):
        # TODO Iterable和six.string_types的判断抽出来放在一个模块，做为Iterable的判断来使用
        if not isinstance(arr, Iterable) or isinstance(arr, six.string_types):
            # arr必须是可以迭代的对象
            raise TypeError('arr not isinstance of Iterable')

        if not isinstance(arr, np.ndarray):
            if isinstance(arr, pd.DataFrame) or isinstance(arr, pd.Series):
                # 如果是pandas直接拿values
                arr = arr.values
            elif isinstance(arr, dict):
                # 针对dict转换np.array
                arr = np.array(list(arr.values())).T
            else:
                arr = np.array(arr)
        return func(arr, *arg, **kwargs)

    return wrapper


@arr_to_numpy
def stats_namedtuple(arr):
    """
    通过序列构造arr的基础统计信息dict, 被arr_to_numpy装饰，统一输出，且这样使用arr.max(), arr.min()等不需要axis参数区别
    与stats_dict的区别只是返回namedtuple对象
    :param arr: pd.DataFrame or pd.Series or Iterable
    :return: MomentsTuple
                eg:
                    count:504
                    max:286.04
                    min:143.67
                    mean:228.48845238095237
                    std:25.538448192811927
                    skewness:-0.282635248604699
                    kurtosis:0.009313464006726946

    """
    count = arr.shape[0]
    if len(arr.shape) > 1 and arr.shape[1] > 1:
        count = arr.shape[0] * arr.shape[1]

    return MomentsTuple(count, arr.max(), arr.min(), arr.mean(), arr.std(),
                        scs.skew(arr), scs.kurtosis(arr))


@arr_to_pandas
def demean(arr, rolling_window=0):
    """
        去均值化后处理demean, 如果输入的是np.array进行转换为pd.DataFrame处理，
        被arr_to_pandas装饰，统一输出，且这样使用arr.mean()不需要axis参数区别np.array轴向
        eg:
            cc.head()
                        tsla	bidu	noah	sfun	goog	vips	aapl
            2014-07-25	223.57	226.50	15.32	12.11	589.02	21.349	97.67
            2014-07-28	224.82	225.80	16.13	12.45	590.60	21.548	99.02
            2014-07-29	225.01	220.00	16.75	12.22	585.61	21.190	98.38
            2014-07-30	228.92	219.13	16.83	11.78	587.42	21.185	98.15
            2014-07-31	223.30	216.05	16.06	11.47	571.60	20.550	95.60

            ABuStatsUtil.demean(cc.head())

                        tsla	bidu	noah	sfun	goog	vips	aapl
            2014-07-25	-1.554	5.004	-0.898	0.104	4.17	0.1846	-0.094
            2014-07-28	-0.304	4.304	-0.088	0.444	5.75	0.3836	1.256
            2014-07-29	-0.114	-1.496	0.532	0.214	0.76	0.0256	0.616
            2014-07-30	3.796	-2.366	0.612	-0.226	2.57	0.0206	0.386
            2014-07-31	-1.824	-5.446	-0.158	-0.536	-13.25	-0.6144	-2.164

            ABuStatsUtil.demean(cc.head().values)

                0	1	2	3	4	5	6
            0	-1.554	5.004	-0.898	0.104	4.17	0.1846	-0.094
            1	-0.304	4.304	-0.088	0.444	5.75	0.3836	1.256
            2	-0.114	-1.496	0.532	0.214	0.76	0.0256	0.616
            3	3.796	-2.366	0.612	-0.226	2.57	0.0206	0.386
            4	-1.824	-5.446	-0.158	-0.536	-13.25	-0.6144	-2.164

            tsla.head()

            2014-07-25    223.57
            2014-07-28    224.82
            2014-07-29    225.01
            2014-07-30    228.92
            2014-07-31    223.30

            ABuStatsUtil.demean(tsla.head())

            2014-07-25   -1.554
            2014-07-28   -0.304
            2014-07-29   -0.114
            2014-07-30    3.796
            2014-07-31   -1.824

            ABuStatsUtil.demean(tsla.head().values)

                0
            0	-1.554
            1	-0.304
            2	-0.114
            3	3.796
            4	-1.824

    :param arr: pd.DataFrame or pd.Series or Iterable
    :param rolling_window: 默认＝0，即不使用移动平均做去均值，rolling_window > 0 生效，
                           注意如果rolling_window值过小将导致去均值后不连续，比如5日，10日的
                           结果只能类似close pct_change的结果，如果需求要钝化，最好是两个月以上
                           的交易日数量，user要根据需求，选择使用的参数，
    :param show: 是否可视化去均值后的结果，默认False
    :return:
    """

    if rolling_window > 0:
        # arr_to_pandas装饰器保证了进来的类型不是pd.DataFrame就是pd.Series
        arr_mean = pd_rolling_mean(arr, window=rolling_window, min_periods=1)
        # arr_mean.fillna(method='bfill', inplace=True)
    else:
        arr_mean = arr.mean()

    demean_v = arr - arr_mean
    return demean_v


def euclidean_distance_matrix(df, scale_end=True, to_similar=False):
    """
    欧式距离(L2范数): 与euclidean_distance_xy的区别主要是，非两两distance计算，只有一个矩阵的输入，
    且输入必须为pd.DataFrame or np.array or 多层迭代序列[[],[]], 注意需要理解数据的测距目的来分析
    是否需要进行scale_start，进行和不进行scale_start的结果将完全不一样，在功能需求及数据理解的情况下
    选择是否进行scale_start

            input:

                        tsla	bidu	noah	sfun	goog	vips	aapl
            2014-07-25	223.57	226.50	15.32	12.110	589.02	21.349	97.67
            2014-07-28	224.82	225.80	16.13	12.450	590.60	21.548	99.02
            2014-07-29	225.01	220.00	16.75	12.220	585.61	21.190	98.38
            ...	...	...	...	...	...	...	...
            2016-07-22	222.27	160.88	25.50	4.850	742.74	13.510	98.66
            2016-07-25	230.01	160.25	25.57	4.790	739.77	13.390	97.34
            2016-07-26	225.93	163.09	24.75	4.945	740.92	13.655	97.76

            ABuStatsUtil.euclidean_distance_matrix(cc, scale_start=True)

            output:

                    tsla	bidu	noah	sfun	goog	vips	aapl
            tsla	0.0000	0.4086	0.7539	0.7942	0.4810	0.7638	0.3713
            bidu	0.4086	0.0000	0.7732	0.7047	0.6185	0.6161	0.4184
            noah	0.7539	0.7732	0.0000	0.7790	0.7174	0.6957	0.7425
            sfun	0.7942	0.7047	0.7790	0.0000	0.9950	0.5422	0.9558
            goog	0.4810	0.6185	0.7174	0.9950	0.0000	1.0000	0.5379
            vips	0.7638	0.6161	0.6957	0.5422	1.0000	0.0000	0.7348
            aapl	0.3713	0.4184	0.7425	0.9558	0.5379	0.7348	0.0000


            ABuStatsUtil.euclidean_distance_matrix(cc, scale_start=False)

                    tsla	bidu	noah	sfun	goog	vips	aapl
            tsla	0.0000	0.0781	0.3314	0.3573	0.6527	0.3386	0.1933
            bidu	0.0781	0.0000	0.2764	0.3018	0.7112	0.2827	0.1392
            noah	0.3314	0.2764	0.0000	0.0284	0.9732	0.0140	0.1408
            sfun	0.3573	0.3018	0.0284	0.0000	1.0000	0.0203	0.1674
            goog	0.6527	0.7112	0.9732	1.0000	0.0000	0.9820	0.8369
            vips	0.3386	0.2827	0.0140	0.0203	0.9820	0.0000	0.1481
            aapl	0.1933	0.1392	0.1408	0.1674	0.8369	0.1481	0.0000

    :param df: pd.DataFrame or np.array or 多层迭代序列[[],[]], 之所以叫df，是因为在内部会统一转换为pd.DataFrame
    :param scale_end: 对结果矩阵进行标准化处理
    :param to_similar: 是否进行后置输出转换similar值
    :return: distance_df，pd.DataFrame对象
    """
    return _distance_matrix(euclidean_distances, df, scale_end, to_similar)


def manhattan_distance_matrix(df, scale_end=True, to_similar=False):
    """
    曼哈顿距离(L1范数): 与manhattan_distances_xy的区别主要是，非两两distance计算，只有一个矩阵的输入，
    且输入必须为pd.DataFrame or np.array or 多层迭代序列[[],[]]，注意需要理解数据的测距目的来分析
    是否需要进行scale_start，进行和不进行scale_start的结果将完全不一样，在功能需求及数据理解的情况下
    选择是否进行scale_start

        eg:
            input:

                        tsla	bidu	noah	sfun	goog	vips	aapl
            2014-07-25	223.57	226.50	15.32	12.110	589.02	21.349	97.67
            2014-07-28	224.82	225.80	16.13	12.450	590.60	21.548	99.02
            2014-07-29	225.01	220.00	16.75	12.220	585.61	21.190	98.38
            ...	...	...	...	...	...	...	...
            2016-07-22	222.27	160.88	25.50	4.850	742.74	13.510	98.66
            2016-07-25	230.01	160.25	25.57	4.790	739.77	13.390	97.34
            2016-07-26	225.93	163.09	24.75	4.945	740.92	13.655	97.76

            ABuStatsUtil.manhattan_distance_matrix(cc, scale_start=True)

            output:

                    tsla	bidu	noah	sfun	goog	vips	aapl
            tsla	0.0000	0.3698	0.6452	0.7917	0.4670	0.7426	0.3198
            bidu	0.3698	0.0000	0.5969	0.7056	0.6495	0.5822	0.4000
            noah	0.6452	0.5969	0.0000	0.7422	0.7441	0.6913	0.6896
            sfun	0.7917	0.7056	0.7422	0.0000	0.9236	0.4489	1.0000
            goog	0.4670	0.6495	0.7441	0.9236	0.0000	0.8925	0.5134
            vips	0.7426	0.5822	0.6913	0.4489	0.8925	0.0000	0.7038
            aapl	0.3198	0.4000	0.6896	1.0000	0.5134	0.7038	0.0000


            ABuStatsUtil.manhattan_distance_matrix(cc, scale_start=False)

            output:

                    tsla	bidu	noah	sfun	goog	vips	aapl
            tsla	0.0000	0.0640	0.3318	0.3585	0.6415	0.3395	0.1906
            bidu	0.0640	0.0000	0.2750	0.3018	0.6982	0.2827	0.1338
            noah	0.3318	0.2750	0.0000	0.0267	0.9733	0.0124	0.1412
            sfun	0.3585	0.3018	0.0267	0.0000	1.0000	0.0191	0.1680
            goog	0.6415	0.6982	0.9733	1.0000	0.0000	0.9809	0.8320
            vips	0.3395	0.2827	0.0124	0.0191	0.9809	0.0000	0.1489
            aapl	0.1906	0.1338	0.1412	0.1680	0.8320	0.1489	0.000

    :param df: pd.DataFrame or np.array or 多层迭代序列[[],[]], 之所以叫df，是因为在内部会统一转换为pd.DataFrame
    :param scale_end: 对结果矩阵进行标准化处理
    :param to_similar: 是否进行后置输出转换similar值
    :return: distance_df，pd.DataFrame对象
    """
    return _distance_matrix(manhattan_distances, df, scale_end, to_similar)


def cosine_distance_matrix(df, scale_end=True, to_similar=False):
    """
    余弦距离: 与cosine_distances_xy的区别主要是，非两两distance计算，只有一个矩阵的输入，
    且输入必须为pd.DataFrame or np.array or 多层迭代序列[[],[]]，注意需要理解数据的测距目的来分析
    是否需要进行scale_start，进行和不进行scale_start的结果将完全不一样，在功能需求及数据理解的情况下
    选择是否进行scale_start

        eg:
            input:

                        tsla	bidu	noah	sfun	goog	vips	aapl
            2014-07-25	223.57	226.50	15.32	12.110	589.02	21.349	97.67
            2014-07-28	224.82	225.80	16.13	12.450	590.60	21.548	99.02
            2014-07-29	225.01	220.00	16.75	12.220	585.61	21.190	98.38
            ...	...	...	...	...	...	...	...
            2016-07-22	222.27	160.88	25.50	4.850	742.74	13.510	98.66
            2016-07-25	230.01	160.25	25.57	4.790	739.77	13.390	97.34
            2016-07-26	225.93	163.09	24.75	4.945	740.92	13.655	97.76


            ABuStatsUtil.cosine_distance_matrix(cc, scale_start=True)

            output:

                    tsla	bidu	noah	sfun	goog	vips	aapl
            tsla	0.0000	0.1743	0.4434	0.2945	0.2394	0.4763	0.1266
            bidu	0.1743	0.0000	0.5808	0.2385	0.3986	0.3034	0.1470
            noah	0.4434	0.5808	0.0000	1.0000	0.3411	0.7626	0.2632
            sfun	0.2945	0.2385	1.0000	0.0000	0.7494	0.4448	0.4590
            goog	0.2394	0.3986	0.3411	0.7494	0.0000	0.9717	0.2806
            vips	0.4763	0.3034	0.7626	0.4448	0.9717	0.0000	0.2669
            aapl	0.1266	0.1470	0.2632	0.4590	0.2806	0.2669	0.0000


            ABuStatsUtil.cosine_distance_matrix(cc, scale_start=False)

            output:

                    tsla	bidu	noah	sfun	goog	vips	aapl
            tsla	0.0000	0.1743	0.4434	0.2945	0.2394	0.4763	0.1266
            bidu	0.1743	0.0000	0.5808	0.2385	0.3986	0.3034	0.1470
            noah	0.4434	0.5808	0.0000	1.0000	0.3411	0.7626	0.2632
            sfun	0.2945	0.2385	1.0000	0.0000	0.7494	0.4448	0.4590
            goog	0.2394	0.3986	0.3411	0.7494	0.0000	0.9717	0.2806
            vips	0.4763	0.3034	0.7626	0.4448	0.9717	0.0000	0.2669
            aapl	0.1266	0.1470	0.2632	0.4590	0.2806	0.2669	0.0000

    :param df: pd.DataFrame or np.array or 多层迭代序列[[],[]], 之所以叫df，是因为在内部会统一转换为pd.DataFrame
    :param scale_end: 对结果矩阵进行标准化处理
    :param to_similar: 是否进行后置输出转换similar值
    :return: distance_df，pd.DataFrame对象
    """
    return _distance_matrix(cosine_distances, df, scale_end, to_similar)


# -*- encoding:utf-8 -*-
"""
    技术线内部执行模块
"""
import math
from enum import Enum
import numpy as np
import pandas as pd
from scipy import optimize as sco, stats
from sklearn.cluster import KMeans
from sklearn.linear_model import LinearRegression
from ultron.utilities.logger import kd_logger
from ultron.ump.core.fixes import xrange
from ultron.kdutils.progress import Progress
from ultron.kdutils import regression
from ultron.ump.core.helper import pd_resample
from ultron.ump.similar.corrcoef import ECoreCorrType, corr_xy

pd.options.display.precision = 4
# numpy精度浮点数显示4位，不使用科学计数法
np.set_printoptions(precision=4, suppress=True)


class ESkeletonHow(Enum):
    """计算骨架走势使用的how"""
    """使用最小值取骨架点位"""
    skeleton_min = 0
    """使用最大值取骨架点位"""
    skeleton_max = 1
    """使用平均值取骨架点位"""
    skeleton_mean = 2
    """使用中位数取骨架点位"""
    skeleton_median = 3
    """使用时间序列最后的元素取骨架点位"""
    skeleton_close = 4
    """
        使用三角模式，即最高，最低，第三点：
        确定取最大值，最小值，第三个点位how_func提供，
        如果np.argmax(arr) > np.argmin(arr)即最大值位置在最小值前面，
        第三点取序列起点，否则取序列终点
    """
    skeleton_triangle = 100


class EShiftDistanceHow(Enum):
    """计算位移路程比的how"""
    """
        使用时间序列最后的元素做为路程的计算基础：
        对应序列的最后一个点位值，标准路程点位值定义
    """
    shift_distance_close = 0
    """
        使用极限值做为路程的计算基础：
        如果p_arr[0] > p_arr[-1]，使用np.min(p_arr)，否则np.max(p_arr)，即上升趋势取max，下跌趋势取min
    """
    shift_distance_maxmin = 1
    """
        使用序列的sum和极限值为路程的计算基础：
        如果abs(p_arr.max() - p_arr[-1]) > abs(p_arr[-1] - p_arr.min()) 取np.min(p_arr)否则np.max(p_arr)
    """
    shift_distance_sum_maxmin = 2


def bfgs_min_pos(find_min_pos, y_len, linear_interp):
    """
    通过scipy.interpolate.interp1d插值形成的模型，通过sco.fmin_bfgs计算min
    :param find_min_pos: 寻找min的点位值
    :param y_len: 原始序列长度，int
    :param linear_interp: scipy.interpolate.interp1d插值形成的模型
    :return: sco.fmin_bfgs成功找到的值，所有失败的或者异常都返回－1
    """
    try:
        local_min_pos = sco.fmin_bfgs(linear_interp, find_min_pos,
                                      disp=False)[0]
    except:
        # 所有失败的或者异常都返回－1
        local_min_pos = -1
    if local_min_pos < 0 or local_min_pos > y_len:
        # 所有失败的或者异常都返回－1
        local_min_pos = -1
    return local_min_pos


def calc_kl_speed(kl, resample=5):
    """
    计算曲线跟随趋势的速度，由于速度是相对的，所以需要在相同周期内与参数曲线进行比对
    :param kl: pd.Series或者numpy序列
    :param resample: 计算速度数值的重采样周期默认5
    :return: 趋势变化敏感速度
    """
    kl_rp = pd_resample(kl, '{}D'.format(resample), how='mean')
    """
        eg: kl_rp
        2011-07-28    1.0027
        2011-08-02    0.9767
        2011-08-07    0.9189
        2011-08-12    0.9174
        2011-08-17    0.9156
    """
    kl_diff = kl_rp.diff()
    """
        eg:
        2011-08-02   -0.0260
        2011-08-07   -0.0578
        2011-08-12   -0.0015
        2011-08-17   -0.0018
    """
    # 二值化 1 -1, 做为速度参数序列慢线
    # noinspection PyTypeChecker
    kl_trend_slow = pd.Series(np.where(kl_diff > 0, 1, -1))
    """
        eg: kl_trend_fast
            0     -1
            1     -1
            2     -1
            3     -1
            4     -1
            5      1
            6      1
            7     -1
            8     -1
            9     -1
    """
    # 慢线向前错一个周期形成快线
    kl_trend_fast = kl_trend_slow.shift(-1)
    """
        egL kl_trend_slow
            0     -1.0
            1     -1.0
            2     -1.0
            3     -1.0
            4      1.0
            5      1.0
            6     -1.0
            7     -1.0
            8     -1.0
            9     -1.0
    """
    kl_trend_ffs = kl_trend_fast[:-1] * kl_trend_slow[:-1]
    """
        慢线 乘 快线 即符号运算
        eg：kl_trend_ffs
        0      1.0
        1      1.0
        2      1.0
        3      1.0
        4     -1.0
        5      1.0
        6     -1.0
        7      1.0
        8      1.0
        9      1.0
    """

    # 符号相同的占所有和的比例即为趋势变化敏感速度值，注意如果没有在相同周期内与参数曲线进行比对的曲线，本速度值即无意义
    speed = kl_trend_ffs.value_counts()[1] / kl_trend_ffs.value_counts().sum()
    return speed


def shift_distance_how(how):
    """
    通过how（EShiftDistanceHow），对应出计算算路程终点点位值使用的计算方法
    注意默认使用shift_distance_close对应标准路程点位值定义方法，其它方法对应的
    路程终点点位值使用的计算方法并非得到最准确的'路程位移比'
    :param how: EShiftDistanceHow对象或者callable即外部可自行设置方法，即计算算路程终点点位值使用的计算方法可自定义
    :return: 计算算路程终点点位值使用的计算方法
    """
    if callable(how):
        # 外部可自行设置方法，即计算算路程终点点位值使用的计算方法可自定义
        return how

    if how == EShiftDistanceHow.shift_distance_maxmin:
        # 如果p_arr[0] > p_arr[-1]，使用np.min(p_arr)，否则np.max(p_arr)，即上升趋势取max，下跌趋势取min
        how = lambda p_arr: np.min(p_arr) if p_arr[0] > p_arr[-1] else np.max(
            p_arr)
    elif how == EShiftDistanceHow.shift_distance_close:
        # 对应序列的最后一个点位值，标准路程点位值定义
        how = lambda p_arr: p_arr[-1]
    elif how == EShiftDistanceHow.shift_distance_sum_maxmin:
        # 如果abs(p_arr.max() - p_arr[-1]) > abs(p_arr[-1] - p_arr.min()) 取np.min(p_arr)否则np.max(p_arr)
        # 即最终的点位绝对距离靠近np.min(p_arr)取np.min(p_arr)否则np.max(p_arr)
        how = lambda p_arr: np.min(p_arr) if abs(p_arr.max() - p_arr[-1]) > abs(p_arr[-1] - p_arr.min()) \
            else np.max(p_arr)
    else:
        raise TypeError('how is error how={}'.format(how))
    return how


def shift_distance(arr, how, slice_start=0, slice_end=-1):
    """
    计算序列arr的'位移路程比'，可视化形式为直接三角展示位移，路程直接的关系，返回
    h_distance(三角底边距离), v_distance(三角垂直距离),
    distance(斜边，路程), shift(位移), sd（位移路程比：shift / distance）
    :param arr: numpy array
    :param how: EShiftDistanceHow，计算路程终点点位值使用的计算方法
    :param slice_start: 如果arr是子序列切片，切片start值， int
    :param slice_end: 如果arr是子序列切片，切片end值， int
    :param color: 直角三角边框颜色，str对象 eg：'r', 'red'
    :return: h_distance(三角底边距离), v_distance(三角垂直距离),
            distance(斜边，路程), shift(位移), sd（位移路程比：shift / distance）
    """
    if slice_end == -1:
        # 因为下面还有slice_end - slice_start等操作，所以转换为正切片值
        slice_end = len(arr)
    # 根据slice_start， slice_end做切片
    slice_arr = arr[slice_start:slice_end]

    # 位置0的值即为路程起点点位值
    start_y = slice_arr[0]
    how = shift_distance_how(how)
    # 根据how的方法计算slice_arr出路程终点点位值
    shift_y = how(slice_arr)
    # 直角三角底边distance
    h_distance = slice_end - slice_start
    # 直角三角垂直边distance
    v_distance = abs(shift_y - start_y)
    # 计算斜边值，即路程值
    distance = math.sqrt(h_distance**2 + v_distance**2)

    # 开始计算位移，首先套上pd.Series做diff结果如下eg形式所示
    diff_slice = pd.Series(slice_arr).diff().fillna(value=0)
    """
        eg: diff_slice
            0     0.0000
            1    -3.7984
            2     7.6824
            3     9.0512
            4     4.3459
            5    -7.7679
                   ...
            44   -3.1482
            45   -1.6426
            46   -2.1216
            47    0.4449
            48   -4.6539
            49    7.1520
    """
    # np.abs(diff_slice).sum计算结果即为位移值
    shift = np.abs(diff_slice).sum()
    # 计算出位移路程比sd
    sd = shift / distance

    # h_distance(三角底边距离), v_distance(三角垂直距离),distance(斜边路程),shift(位移),sd（位移路程比：shift / distance）
    return h_distance, v_distance, distance, shift, sd


def calc_pair_speed(kl,
                    benchmark_kl,
                    kl_close=None,
                    benchmark_close=None,
                    resample=5):
    # 通过calc_kl_speed计算趋势跟随速度
    kl_speed = calc_kl_speed(kl, resample)
    benchmark_kl_speed = calc_kl_speed(benchmark_kl, resample)
    # 两个走势的SPERM相关性
    kl_close = kl if kl_close is None else kl_close
    benchmark_close = benchmark_kl if benchmark_close is None else benchmark_close
    corr = corr_xy(kl_close, benchmark_close, ECoreCorrType.E_CORE_TYPE_SPERM)
    # 返回参数symbol, benchmark_symbol所对应的趋势变化敏感速度数值, 以及相关性＊敏感度＝敏感度置信度
    return kl_speed, benchmark_kl_speed, corr


def support_resistance_pos(x, support_resistance_y, best_poly=0, label=None):
    """
    分析获取序列阻力位或者支撑位，通过sco.fmin_bfgs寻找阻力位支撑位，阻力位点也是通过sco.fmin_bfgs寻找，
    但是要求传递进来的序列已经是标准化后取反的序列
    eg：
        demean_y = ABuStatsUtil.demean(self.tl)： 首先通过demean将序列去均值
        resistance_y = demean_y * -1 ：阻力位序列要取反
        support_y = demean_y ：支持位序列不需要取反
    sco.fmin_bfgs使用的模型函数为polynomial.Chebyshev多项拟合函数，poly的次数确定
    由ABuRegUtil.search_best_poly得到，即best_poly次多项式拟合回归的趋势曲线可以比较完美的代表原始曲线y的走势，
    为了得到更多的阻力支持位种子点位值，使用：
        np.polynomial.Chebyshev.fit(x, support_resistance_y, best_poly * g_upport_resistance_unit)
        g_upport_resistance_unit默认＝3
    best_poly * 3，即将poly次数又扩大了3倍，可以改变g_upport_resistance_unit获取更多的阻力位支撑位种子点，但
    速度会更慢。
    :param x: 待分析的序列np.array
    :param support_resistance_y:
    :param best_poly: 函数使用者可设置best_poly, 设置后就不使用ABuRegUtil.search_best_poly寻找了
    :param label: 进度条显示的等待文字，str对象
    :return: 阻力位或者支撑位序列
    """
    if best_poly <= 0:
        # 由ABuRegUtil.search_best_poly得到，即best_poly次多项式拟合回归的趋势曲线可以比较完美的代表原始曲线y的走势
        # 注意这里poly_min＝7，即从7次ploy开始，提升效率
        best_poly = regression.search_best_poly(support_resistance_y,
                                                poly_min=7,
                                                zoom=False,
                                                only_poly=True)
        # 根据曲线的长度倍乘best_poly
        best_poly *= int(math.ceil(len(x) / 120))
    best_poly = int(len(x) / 20) if best_poly < int(len(x) / 20) else best_poly

    # 为了得到更多的阻力支持位种子点位值->best_poly * 3，即将poly次数又扩大了3倍
    p = np.polynomial.Chebyshev.fit(x, support_resistance_y, best_poly)

    # 需要使用set，因为需要过滤重复的
    support_resistance = set()
    # 属于耗时操作，构建进度条显示
    with Progress(len(support_resistance_y), 0, label) as progess:
        for index in xrange(0, len(support_resistance_y), 1):
            progess.show(index + 1)
            local_min_pos = int(
                bfgs_min_pos(index, len(support_resistance_y), p))
            if local_min_pos == -1:
                # 其实主要就是利用这里找不到的情况进行过滤
                continue
            # 将local_min_pos加到集合中
            support_resistance.add(local_min_pos)
    support_resistance = list(support_resistance)
    return support_resistance


def select_k_support_resistance(support_resistance, thresh=0.06):
    """
    对阻力位或者支撑位序列从1-序列个数开始聚类，多个聚类器的方差值进行比较，
    通过方差阀值等方法找到最佳聚类个数，最终得到kmean最佳分类器对象
    :param support_resistance: 阻力位或者支撑位序列
    :param thresh: 聚类方差比例阀值，默认0.06
    :param label: 可视化显示的label，主要用来区别阻力位和支撑位的聚类分析结果
    :return: 最佳分类器对象，KMeans类型
    """
    # 阻力位或者支撑位序列从1-序列个数开始聚类
    k_rng = xrange(1, len(support_resistance))
    # 聚类里某一个类别的最低点只有一个是有效
    est_arr = [KMeans(n_clusters=k).fit(support_resistance) for k in k_rng]
    sum_squares = [e.inertia_ for e in est_arr]
    """
        eg: sum_squares 形如
        [568824.24742527946, 63692.462671013389, 23023.512755456246, 11106.460471044047, 4739.3661037023803, 2
        479.206823270833, 1640.5275739375002, 1095.9099614375, 589.28019737500006, 160.28686404166669,
        45.229672916666679, 31.689656250000013, 18.94814375000001, 10.944315625000012, 5.5643156249999999,
        0.6365031249999995]
    """
    # 各类别相对中心点比， 中心点位1类
    diff_squares = [squares / sum_squares[0] for squares in sum_squares]
    diff_squares_pd = pd.Series(diff_squares)

    thresh_pd = diff_squares_pd[diff_squares_pd < thresh]
    if thresh_pd.shape[0] > 0:
        """
            eg：thresh_pd 形如index[0]代表最小的聚类个数值
            2     4.0476e-02
            3     1.9525e-02
            4     8.3319e-03
            5     4.3585e-03
            6     2.8841e-03
            7     1.9266e-03
            8     1.0360e-03
            9     2.8179e-04
        """
        select_k = k_rng[thresh_pd.index[0]]
    else:
        # 没有符合的，就安装最多的个数聚类
        select_k = k_rng[-1]

    # 通过select_k取代分类器
    est = est_arr[select_k - 1]
    return est


def support_resistance_predict(x, y, est, support_resistance, is_support):
    """
    通过最优聚类器est从阻力位或者支撑位序列support_resistance中的序列进行聚类predict
    对聚类结果进行可视化，最终从每个cluster中找到唯一值代表这个分类，如果是支撑位要找到最小值的index
    如果是阻力位找到最大值的index
    :param x: 可迭代序列x
    :param y: 可迭代序列y
    :param est: 最佳分类器对象，KMeans类型
    :param support_resistance: 阻力位或者支撑位序列
    :param is_support: 是否是进行支撑位support_resistance_predict
    :return: 返回从每个cluster中找到唯一值代表这个分类的x值组成的序列。list
    """
    support_resistance_k = est.predict(support_resistance)
    """
        eg: support_resistance_k形如：
        array([0, 0, 1, 1, 0, 1, 1, 0, 0, 2, 1, 0, 0, 2, 0, 0, 0], dtype=int32)
    """
    d_pd = pd.DataFrame(support_resistance, columns=['x', 'y'])
    d_pd['cluster'] = support_resistance_k
    """
        eg：d_pd形如：
                x        y  cluster
        0     0.0  15.5400        0
        1    97.0  21.3475        0
        2   482.0  22.9625        1
        3   387.0  21.6725        1
        4     5.0  16.2350        0
        5   426.0  23.7650        1
        6   459.0  23.2800        1
        7    48.0  13.1475        0
        8    49.0  13.6700        0
        9   283.0  20.9150        2
        10  497.0  25.2240        1
        11   53.0  13.6537        0
        12   23.0  15.3150        0
        13  280.0  19.9900        2
        14  121.0  18.3475        0
        15   27.0  15.2275        0
        16   95.0  23.9475        0
    """
    k_list = list()

    # 最终从每个cluster中找到唯一值代表这个分类
    for k in set(support_resistance_k):
        if is_support:
            # 如果是支撑位要找到最小值的index
            #ind = d_pd[d_pd['cluster'] == k]['y'].argmin()  # argmin 函数升级
            cluster_pd = d_pd[d_pd['cluster'] == k].copy()
            ind = cluster_pd['y'].argmin()
            choice_x = int(cluster_pd.iloc[ind]['x'])
        else:
            # 如果是阻力位找到最大值的index
            #ind = d_pd[d_pd['cluster'] == k]['y'].argmax()
            cluster_pd = d_pd[d_pd['cluster'] == k].copy()
            ind = cluster_pd['y'].argmax()
            choice_x = int(cluster_pd.iloc[ind]['x'])
        # 最终的结果只要x的值
        #choice_x = int(d_pd.iloc[ind]['x'])
        k_list.append(choice_x)
    # 返回从每个cluster中找到唯一值代表这个分类的x值组成的序列
    return k_list


def create_support_resistance_trend(x, y, trend_pos, label, only_last=False):
    """
    通过trend_pos绘制阻力线或者支撑线，only_last控制只绘制时间序列中最后一个发现的阻力或支撑，
    plot_org控制是否绘制线段还是直线，plot_org＝True时绘制线段，否则通过LinearRegression进行
    直线绘制
    :param x: 可迭代序列x
    :param y: 可迭代序列y
    :param trend_pos: 趋势点x序列，绘制时需要通过原始序列y获取y[x]值
    :param label: 可视化显示的文字，用来区分阻力位和支撑位
    :param only_last: 控制只绘制时间序列中最后一个发现的阻力或支撑
    """
    # trend_pos类型为list容器
    trend_pos.sort()
    if len(trend_pos) < 2:
        kd_logger.info('{} len(trend_pos) < 2 !'.format(label))
        return

    if only_last:
        # 只绘制时间序列中最后一个发现的阻力或支撑，因为前面trend_pos.sort()了，即序列中最后两个元素
        trend_pos = trend_pos[-2:]

    y_trend_arr = []
    for ind, trend_start in enumerate(trend_pos):
        if ind == len(trend_pos) - 1:
            continue
        trend_end = trend_pos[ind + 1]

        x_org = [trend_start, trend_end]
        y_org = [y[trend_start], y[trend_end]]

        # 通过LinearRegression学习线段
        reg = LinearRegression()
        reg.fit(np.array(x_org).reshape(-1, 1), np.array(y_org).reshape(-1, 1))

        x_line = [x[0], x[-1]]
        # predict序列的第一个和最后一个点，即将线段变成了直线，延伸了阻力位，支撑位
        y_line = reg.predict(np.array(x_line).reshape(-1, 1)).reshape(-1, )
        # 把端点结果返回给外面
        y_trend_arr.append(y_line)
    return y_trend_arr


def regress_trend_channel(arr):
    """
    通过arr计算拟合曲线及上下拟合通道曲线，返回三条拟合曲线，组成拟合通道
    :param arr: numpy array
    :return: y_below, y_fit, y_above
    """
    # 通过ABuRegUtil.regress_y计算拟合曲线和模型reg_mode，不使用缩放参数zoom
    reg_mode, y_fit = regression.regress_y(arr, zoom=False)
    reg_params = reg_mode.params

    x = np.arange(0, len(arr))
    a = reg_params[0]
    b = reg_params[1]
    # 通过argmin寻找出原始序列和拟合序列差值的最小点，差值最小代表点位离拟合曲线远，eg: 100 - 80 < 100 - 90
    min_ind = (arr.T - y_fit).argmin()
    # 根据a, b计算出below值, 注意这里是差，eg: below：100 － 80 ＝ 20
    below = x[min_ind] * b + a - arr[min_ind]
    # 计算x * b + a但- below，即拟合曲线保持相同的斜率整体下移below值
    y_below = x * b + a - below

    # 通过argmax寻找出原始序列和拟合序列差值的最大点，差值最小代表点位离拟合曲线远，eg: 120 - 100 > 120 - 110
    max_ind = (arr.T - y_fit).argmax()
    # 根据a, b计算出above值, 注意这里是差，eg: above 100 - 120 ＝ -20, 即above是负数
    above = x[max_ind] * b + a - arr[max_ind]
    # 计算x * b + a但整天- above，由于above是负数，即相加 即拟合曲线保持相同的斜率整体上移above值
    y_above = x * b + a - above
    return y_below, y_fit, y_above


def find_percent_point(percents, y):
    """
    技术线比例分割的区域, 针对输入的比例迭代操作后
    分别使用stats.scoreatpercentile和 (y.max() - y.min()) * pt + y.min()两种
    方式进行计算的分割值, 返回对象为比例值为key的字典对象
    eg:
        input:
            percents = (0.1, 0.9)
        output:
            {0.1: (15.732749999999999, 15.5075), 0.9: (31.995000000000005, 34.387500000000003)}

    :param percents: 可迭代序列，eg: (0.1, 0.9), [0.3, 0,4, 0.8]
    :param y: 计算分割线的序列
    :return: 比例值为key的字典对象
    """
    percent_point_dict = {
        pt: (stats.scoreatpercentile(y, np.round(pt * 100, 1)),
             (y.max() - y.min()) * pt + y.min())
        for pt in percents
    }

    return percent_point_dict


def find_golden_point_ex(x, y):
    sp382 = stats.scoreatpercentile(y, 38.2)
    sp618 = stats.scoreatpercentile(y, 61.8)
    sp50 = stats.scoreatpercentile(y, 50.0)
    return sp382, sp50, sp618


def find_golden_point(x, y):
    """视觉黄金分割计算方法"""

    cs_max = y.max()
    cs_min = y.min()

    sp382 = (cs_max - cs_min) * 0.382 + cs_min
    sp618 = (cs_max - cs_min) * 0.618 + cs_min
    sp50 = (cs_max - cs_min) * 0.5 + cs_min
    return sp382, sp618, sp50


def below_above_gen(x, y):
    """
    (y, x) if x > y else (x, y)
    :param x: 支持比较操作的对象
    :param y: 支持比较操作的对象
    """
    return (y, x) if x > y else (x, y)


def skeleton_how(how):
    """
    根据how映射计算数据序列骨架点位的方法
    :param how: ESkeletonHow对象或者callable即外部可自行设置方法，即计算数据序列骨架点位的方法可自定义
    :return:
    """
    if callable(how):
        # callable即外部可自行设置方法
        return how

    if how == ESkeletonHow.skeleton_min:
        how_func = np.min
    elif how == ESkeletonHow.skeleton_max:
        how_func = np.max
    elif how == ESkeletonHow.skeleton_mean:
        how_func = np.mean
    elif how == ESkeletonHow.skeleton_median:
        how_func = np.median
    elif how == ESkeletonHow.skeleton_close:
        # 取序列最后一个元素做为采样骨架点位
        how_func = lambda arr: arr[-1]
    elif how == ESkeletonHow.skeleton_triangle:
        # 三角模式骨架点位：确定取最大值，最小值，第三个点位how_func提供
        # 如果np.argmax(arr) > np.argmin(arr)即最大值位置在最小值前面，第三点取序列起点，否则取序列终点
        how_func = lambda arr, start: (start, arr[0]) if np.argmax(arr) > np.argmin(arr) else \
            (len(arr) + start, arr[-1])
    else:
        raise TypeError('how is error how={}'.format(how))
    return how_func


# -*- coding: utf-8 -*-

from datetime import datetime
import numbers,json

import numpy as np
from joblib import cpu_count

def check_random_state(seed):
    """Turn seed into a np.random.RandomState instance

    Parameters
    ----------
    seed : None | int | instance of RandomState
        If seed is None, return the RandomState singleton used by np.random.
        If seed is an int, return a new RandomState instance seeded with seed.
        If seed is already a RandomState instance, return it.
        Otherwise raise ValueError.

    """
    if seed is None or seed is np.random:
        return np.random.mtrand._rand
    if isinstance(seed, (numbers.Integral, np.integer)):
        return np.random.RandomState(seed)
    if isinstance(seed, np.random.RandomState):
        return seed
    raise ValueError('%r cannot be used to seed a numpy.random.RandomState'
                     ' instance' % seed)

class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, datetime.datetime):
            return obj.strftime('%Y-%m-%d %H:%M:%S')
        else:
            return super(NumpyEncoder, self).default(obj)


# -*- encoding:utf-8 -*-
"""
    全局环境配置模块
"""

from enum import Enum
import os
"""主进程pid，使用并行时由于EnvProcess会拷贝主进程注册了的模块信息，所以可以用g_main_pid来判断是否在主进程"""
g_main_pid = os.getpid()
g_is_ipython = True

g_is_mac_os = False
try:
    # noinspection PyUnresolvedReferences
    __IPYTHON__
except NameError:
    g_is_ipython = False


# TODO 缩短 E_MARKET_TARGET_US－>US
class EMarketTargetType(Enum):
    """
        交易品种类型，即市场类型，
        eg. 美股市场, A股市场, 港股市场, 国内期货市场,
            美股期权市场, TC币市场（比特币等
    """
    """美股市场"""
    E_MARKET_TARGET_US = 'us'
    """A股市场"""
    E_MARKET_TARGET_CN = 'hs'
    """港股市场"""
    E_MARKET_TARGET_HK = 'hk'
    """国内期货市场"""
    E_MARKET_TARGET_FUTURES_CN = 'futures_cn'
    """国际期货市场"""
    E_MARKET_TARGET_FUTURES_GLOBAL = 'futures_global'
    """美股期权市场"""
    E_MARKET_TARGET_OPTIONS_US = 'options_us'
    """TC币市场（比特币等）"""
    E_MARKET_TARGET_TC = 'tc'


g_market_trade_year = 250


# TODO 缩短 E_DATA_FETCH_NORMAL－>NORMAL
class EMarketDataFetchMode(Enum):
    """
        金融时间数据获取模式
    """
    """普通模式，尽量从本地获取数据，本地数据不满足的情况下进行网络请求"""
    E_DATA_FETCH_NORMAL = 0
    """强制从本地获取数据，本地数据不满足的情况下，返回None"""
    E_DATA_FETCH_FORCE_LOCAL = 1
    """强制从网络获取数据，不管本地数据是否满足"""
    E_DATA_FETCH_FORCE_NET = 2


class EDataCacheType(Enum):
    """
        金融时间序列数据缓存类型
    """
    """读取及写入最快 但非固态硬盘写入慢，存贮空间需要大"""
    E_DATA_CACHE_HDF5 = 0
    """读取及写入最慢 但非固态硬盘写速度还可以，存贮空间需要小"""
    E_DATA_CACHE_CSV = 1
    """适合分布式扩展，存贮空间需要大"""
    E_DATA_CACHE_MONGODB = 2


# TODO EMarketDataSplitMode移动到市场请求相关对应的模块中
class EMarketDataSplitMode(Enum):
    """
        ABuSymbolPd中请求参数，关于是否需要与基准数据对齐切割
    """
    """直接取出所有data，不切割，即外部需要切割"""
    E_DATA_SPLIT_UNDO = 0
    """内部根据start，end取切割data"""
    E_DATA_SPLIT_SE = 1


class EMarketSubType(Enum):
    """
        子市场（交易所）类型定义
    """
    """美股纽交所NYSE"""
    US_N = 'NYSE'
    """美股纳斯达克NASDAQ"""
    US_OQ = 'NASDAQ'
    """美股粉单市场"""
    US_PINK = 'PINK'
    """美股OTCMKTS"""
    US_OTC = 'OTCMKTS'
    """美国证券交易所"""
    US_AMEX = 'AMEX'
    """未上市"""
    US_PREIPO = 'PREIPO'
    """港股hk"""
    HK = 'hk'
    """上证交易所sh"""
    SH = 'sh'
    """深圳交易所sz"""
    SZ = 'sz'
    """大连商品交易所DCE'"""
    DCE = 'DCE'
    """郑州商品交易所ZZCE'"""
    ZZCE = 'ZZCE'
    """上海期货交易所SHFE'"""
    SHFE = 'SHFE'
    """伦敦金属交易所"""
    LME = 'LME'
    """芝加哥商品交易所"""
    CBOT = 'CBOT'
    """纽约商品交易所"""
    NYMEX = 'NYMEX'
    """币类子市场COIN'"""
    COIN = 'COIN'


"""切换目标操作市场，美股，A股，港股，期货，比特币等，默认美股市场"""
g_market_target = EMarketTargetType.E_MARKET_TARGET_US

try:
    # noinspection PyUnresolvedReferences
    import psutil
    """有psutil，使用psutil.cpu_count计算cpu个数"""
    g_cpu_cnt = psutil.cpu_count(logical=True) * 1
except ImportError:
    if True:
        # noinspection PyUnresolvedReferences
        g_cpu_cnt = os.cpu_count()
    else:
        import multiprocessing as mp

        g_cpu_cnt = mp.cpu_count()
except:
    # 获取cpu个数失败，默认4个
    g_cpu_cnt = 4

# ＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊主裁 start ＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊
# TODO 内置ump的设置move到UltronUmpManager中
"""是否开启裁判拦截机制: 主裁deg，默认关闭False"""
g_enable_ump_main_deg_block = False
"""是否开启裁判拦截机制: 主裁jump，默认关闭False"""
g_enable_ump_main_jump_block = False
"""是否开启裁判拦截机制: 主裁price，默认关闭False"""
g_enable_ump_main_price_block = False
"""是否开启裁判拦截机制: 主裁wave，默认关闭False"""
g_enable_ump_main_wave_block = False
"""是否开启裁判拦截机制: 边裁deg，默认关闭False"""
g_enable_ump_edge_deg_block = False
"""是否开启裁判拦截机制: 边裁price，默认关闭False"""
g_enable_ump_edge_price_block = False
"""是否开启裁判拦截机制: 边裁wave，默认关闭False"""
g_enable_ump_edge_wave_block = False
"""是否开启裁判拦截机制: 边裁full，默认关闭False"""
g_enable_ump_edge_full_block = False
"""是否开启机器学习特征收集, 开启后速度会慢，默认关闭False"""
g_enable_ml_feature = False
"""是否启用外部用户使用append_user_ump添加的ump对交易进行拦截决策"""
g_enable_user_ump = False


# -*- coding: utf-8 -*-
import time, datetime, pdb
import numpy as np
from joblib import Parallel, delayed
import warnings, time, itertools, copy
from ultron.optimize.geneticist.program import Program
from ultron.optimize.geneticist import program
from ultron.utilities.utils import check_random_state
from ultron.utilities.jobs import partition_estimators

warnings.filterwarnings("ignore")

MAX_INT = np.iinfo(np.int32).max
MIN_INT = np.iinfo(np.int32).min


def parallel_evolve(n_programs, parents, features, X, Y, seeds,
                    greater_is_better, gen, params):
    tournament_size = params['tournament_size']
    model_sets = params['model_sets']
    params_sets = params['params_sets']
    init_method = params['init_method']
    method_probs = params['method_probs']
    p_point_replace = params['p_point_replace']
    fitness = params['fitness']
    mode = params['mode']
    n_splits = params['n_splits']
    custom_params = params['custom_params']

    def _contenders(tour_parents):
        contenders = random_state.randint(0, len(tour_parents), 6)
        return [tour_parents[p] for p in contenders]

    def _tournament(tour_parents, model_name=None):
        if model_name is not None:
            tour_parents = [
                p for p in tour_parents if p._model_name == model_name
            ]
        contenders = random_state.randint(0, len(tour_parents),
                                          tournament_size)
        raw_fitness = [tour_parents[p]._raw_fitness for p in contenders]
        if greater_is_better:
            parent_index = contenders[np.argmax(raw_fitness)]
        else:
            parent_index = contenders[np.argmin(raw_fitness)]
        return tour_parents[parent_index], parent_index

    programs = []
    for i in range(n_programs):
        random_state = check_random_state(seeds[i])
        if parents is None:
            program = None
            genome = None
            p_init_method = init_method
        else:
            method = random_state.uniform()
            parent, parent_index = _tournament(copy.deepcopy(parents))
            ori_parent = copy.deepcopy(parent)

            contenders = _contenders(copy.deepcopy(parents))
            for contender in contenders:
                if contender._model_name != parent._model_name:
                    continue
                program, removed, remains = parent.crossover(
                    contender._program, random_state)
                parent = Program(params_sets=params_sets,
                                 model_sets=model_sets,
                                 random_state=random_state,
                                 method=init_method,
                                 p_point_replace=p_point_replace,
                                 gen=gen,
                                 fitness=fitness,
                                 program=program)
                break
            # 新特征种群
            if random_state.uniform() < method_probs[1]:
                program = Program(params_sets=params_sets,
                                  model_sets=model_sets,
                                  random_state=random_state,
                                  method=init_method,
                                  p_point_replace=p_point_replace,
                                  gen=gen,
                                  fitness=fitness,
                                  model_name=parent._model_name)
                program, removed, remains = parent.crossover(
                    program._program, random_state)
                parent = Program(params_sets=params_sets,
                                 model_sets=model_sets,
                                 random_state=random_state,
                                 method=init_method,
                                 p_point_replace=p_point_replace,
                                 gen=gen,
                                 fitness=fitness,
                                 program=program)

            if method < method_probs[0]:  # crossover
                donor, donor_index = _tournament(copy.deepcopy(parents),
                                                 model_name=parent._model_name)
                program, removed, remains = parent.crossover(
                    donor._program, random_state)
                genome = {
                    'method': 'Crossover',
                    'parent_idx': parent_index,
                    'parent_nodes': removed,
                    'donor_idx': donor_index,
                    'donor_nodes': remains
                }
            elif method < method_probs[1]:  # point_mutation
                program, removed, remains = parent.point_mutation(random_state)
                genome = {
                    'method': 'Point Mutation',
                    'parent_idx': parent_index,
                    'parent_nodes': removed
                }
            else:
                program = parent.reproduce()  # reproduction
                genome = {
                    'method': 'Reproduction',
                    'parent_idx': parent_index,
                    'parent_nodes': []
                }
            # 与原始自身进行交叉
            if random_state.uniform() < method_probs[0]:
                program = Program(params_sets=params_sets,
                                  model_sets=model_sets,
                                  random_state=random_state,
                                  method=init_method,
                                  p_point_replace=p_point_replace,
                                  gen=gen,
                                  fitness=fitness,
                                  program=program)
                program, removed, remains = program.crossover(
                    ori_parent._program, random_state)
            p_init_method = genome['method']
        program = Program(params_sets=params_sets,
                          model_sets=model_sets,
                          random_state=random_state,
                          method=p_init_method,
                          p_point_replace=p_point_replace,
                          gen=gen,
                          fitness=fitness,
                          program=program)
        default_value = MIN_INT if greater_is_better else MAX_INT
        program.raw_fitness(features=features,
                            X=X,
                            Y=Y,
                            mode=mode,
                            default_value=default_value,
                            custom_params=custom_params,
                            n_splits=n_splits)
        programs.append(program)
    return programs


class Gentic(object):

    def __init__(
            self,
            model_sets,
            params_sets,
            population_size=2000,
            generations=MAX_INT,
            tournament_size=20,
            n_jobs=1,
            p_crossover=0.9,
            p_point_mutation=0.01,
            p_subtree_mutation=0.00,
            p_hoist_mutation=0.00,
            p_point_replace=0.8,
            stopping_criteria=0.0,
            fitness=None,
            custom_params=None,
            save_model=None,
            random_state=None,
            greater_is_better=True,  #True 倒序， False 正序
            convergence=None,
            rootid=0,
            standard_score=0.9,  # None代表 根据tournament_size保留种群  standard_score保留种群
            verbose=1):

        self._population_size = population_size
        self._generations = MAX_INT if generations == 0 else generations
        self._tournament_size = tournament_size
        self._stopping_criteria = stopping_criteria
        self.init_method = 'full'
        self._model_sets = model_sets
        self._params_sets = params_sets
        self._p_crossover = p_crossover
        self._p_subtree_mutation = p_subtree_mutation
        self._p_hoist_mutation = p_hoist_mutation
        self._p_point_mutation = p_point_mutation
        self._p_point_replace = p_point_replace
        self._random_state = random_state
        self._fitness = fitness
        self._save_model = save_model
        self._greater_is_better = greater_is_better
        self._standard_score = standard_score
        self._convergence = convergence
        self._custom_params = custom_params
        self._con_time = 0
        self._n_jobs = n_jobs
        self._verbose = verbose
        self._rootid = int(
            time.time() * 1000000 +
            datetime.datetime.now().microsecond) if rootid == 0 else rootid

    def filter_programs(self, population):
        ## 保留符合条件的种群(1.种群有效 2.分数优于基准分 3.符合指定个数)
        valid_prorams = np.array(population)[[
            program._is_valid for program in population
        ]]  # 只保留有效种群
        ## 删除重复种群
        identification_dict = {}
        for program in valid_prorams:
            identification_dict[program._identification] = program
        valid_prorams = list(identification_dict.values())
        fitness = [program._raw_fitness for program in valid_prorams]
        if self._standard_score is not None:  #分数筛选且第二代开始
            if self._greater_is_better:
                best_programs = np.array([
                    program for program in valid_prorams
                    if program._raw_fitness > self._standard_score
                ])
            else:
                best_programs = np.array([
                    program for program in valid_prorams
                    if program._raw_fitness < self._standard_score
                ])

        #若不满足分数，则进行排序选出前_tournament_size
        if len(best_programs
               ) < self._tournament_size or self._standard_score is None:
            if self._greater_is_better:
                best_programs = np.array(valid_prorams)[np.argsort(
                    fitness)[-self._tournament_size:]]
            else:
                best_programs = np.array(valid_prorams)[np.argsort(
                    fitness)[:self._tournament_size]]
        return best_programs

    def train(self, features, X, Y, mode, n_splits):
        random_state = check_random_state(self._random_state)
        self._method_probs = np.array(
            [self._p_crossover, self._p_point_mutation])

        self._method_probs = np.cumsum(self._method_probs)
        if self._method_probs[-1] > 1:
            raise ValueError('The sum of p_crossover, '
                             'and p_point_mutation should '
                             'total to 1.0 or less.')
        params = {}
        params['tournament_size'] = self._tournament_size
        params['init_method'] = self.init_method
        params['model_sets'] = self._model_sets
        params['params_sets'] = self._params_sets
        params['method_probs'] = self._method_probs
        params['p_point_replace'] = self._p_point_replace
        params['fitness'] = self._fitness
        params['mode'] = mode
        params['n_splits'] = n_splits
        params['custom_params'] = self._custom_params

        self._programs = []
        self._best_programs = None
        self._run_details = {
            'generation': [],
            'average_fitness': [],
            'best_fitness': [],
            'generation_time': [],
            'best_programs': []
        }
        prior_generations = len(self._programs)
        n_more_generations = self._generations - prior_generations
        for gen in range(prior_generations, self._generations):
            start_time = time.time()
            if gen == 0:
                parents = None
            else:
                parents = self._programs[gen - 1]
                parents = [parent for parent in parents if parent._is_valid]

            n_jobs, n_programs, starts = partition_estimators(
                self._population_size, self._n_jobs)
            seeds = random_state.randint(MAX_INT, size=self._population_size)
            population = Parallel(n_jobs=n_jobs, verbose=self._verbose)(
                delayed(parallel_evolve)(n_programs[i], parents, features, X,
                                         Y, seeds, self._greater_is_better,
                                         gen, params) for i in range(n_jobs))

            population = list(itertools.chain.from_iterable(population))

            population = [
                program for program in population if program._is_valid
            ]
            if len(population) == 0:
                break

            if self._best_programs is None:
                self._programs.append(population)
            else:
                identification_dict = {}
                valid_prorams = list(
                    np.concatenate([population, self._best_programs]))
                for program in valid_prorams:
                    identification_dict[program._identification] = program
                valid_prorams = list(identification_dict.values())
                self._programs.append(valid_prorams)

            best_programs = self.filter_programs(population)
            if self._best_programs is not None:
                best_programs = np.concatenate(
                    [best_programs, self._best_programs])
                best_programs = self.filter_programs(best_programs)

            self._best_programs = best_programs
            for program in self._best_programs:
                program.log()

            fitness = [program._raw_fitness for program in self._best_programs]
            self._run_details['generation'].append(gen)
            self._run_details['average_fitness'].append(np.mean(fitness))
            generation_time = time.time() - start_time
            self._run_details['generation_time'].append(generation_time)
            self._run_details['best_programs'].append(self._best_programs)
            print(
                'ExpendTime:%f,Generation:%d,Tournament:%d, Fitness Mean:%f,Fitness Max:%f,Fitness Min:%f'
                % (generation_time, gen, len(best_programs), np.mean(fitness),
                   np.max(fitness), np.min(fitness)))

            ##
            if self._save_model is not None:
                self._save_model(gen, self._rootid,
                                 self._run_details['best_programs'][-1])
            if self._greater_is_better:
                best_fitness = fitness[np.argmax(fitness)]
                if best_fitness >= self._stopping_criteria:
                    break
            else:
                best_fitness = fitness[np.argmin(fitness)]
                if best_fitness <= self._stopping_criteria:
                    break

            if np.mean(fitness) == MIN_INT or best_fitness == MIN_INT:
                break
            self._run_details['best_fitness'].append(best_fitness)

            # 收敛值判断
            if self._convergence is None or gen == 0:
                continue
            d_value = np.mean(fitness) - self._run_details['average_fitness'][
                gen - 1]
            print('d_value:%f,convergence:%f,con_time:%d' %
                  (d_value, self._convergence, self._con_time))
            if abs(d_value) < self._convergence:
                self._con_time += 1
                if self._con_time > 5:
                    break
            else:
                self._con_time = 0



# -*- coding: utf-8 -*-
import numpy as np
from ....utilities.singleton import Singleton
from ultron.sentry.Analysis.SecurityValueHolders import SecurityLatestValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityCurrentValueHolder

##横截面变异
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityAverageValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityDiffValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecuritySignValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityExpValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityLogValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecuritySqrtValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityAbsValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityAcosValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityAcoshValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityAsinValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityAsinhValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityNormInvValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityCeilValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityFloorValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityRoundValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecuritySigmoidValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityTanhValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecuritySimpleReturnValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityLogReturnValueHolder

##横截面交叉
from ultron.sentry.Analysis.SecurityValueHolders import SecurityAddedValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecuritySubbedValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityMultipliedValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityDividedValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityLtOperatorValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityLeOperatorValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityGtOperatorValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityGeOperatorValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityEqOperatorValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityNeOperatorValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityAndOperatorValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityOrOperatorValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityMinimumValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityMaximumValueHolder

from ultron.sentry.Analysis.CrossSectionValueHolders import CSAverageSecurityValueHolder
from ultron.sentry.Analysis.CrossSectionValueHolders import CSAverageAdjustedSecurityValueHolder
from ultron.sentry.Analysis.CrossSectionValueHolders import CSResidueSecurityValueHolder
from ultron.sentry.Analysis.CrossSectionValueHolders import CSRankedSecurityValueHolder
from ultron.sentry.Analysis.CrossSectionValueHolders import CSPercentileSecurityValueHolder
from ultron.sentry.Analysis.CrossSectionValueHolders import CSZScoreSecurityValueHolder

## 横截面变异
#from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityPowValueHolder # 指定默认参数

## 时间序列变异
#from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityMACDValueHolder #双默认参数

from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityXAverageValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingAverage
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingDecay
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingMax
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingArgMax
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingMin
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingArgMin
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingRank
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingQuantile
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingAllTrue
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingAnyTrue
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingSum
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingVariance
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingStandardDeviation
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingCountedPositive
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingPositiveAverage
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingPositiveDifferenceAverage
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingNegativeDifferenceAverage
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingRSI
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingLogReturn

from ultron.sentry.Analysis.SecurityValueHolders import SecurityDeltaValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityShiftedValueHolder

## 时间序列交叉
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingCorrelation

from ultron.sentry.api import CSMean
from ultron.sentry.api import CSMeanAdjusted
from ultron.sentry.api import CSRes
from ultron.sentry.api import CSRank
from ultron.sentry.api import CSQuantiles
from ultron.sentry.api import CSZScore

from ultron.sentry.api import ADDED
from ultron.sentry.api import SUBBED
from ultron.sentry.api import MUL
from ultron.sentry.api import DIV
from ultron.sentry.api import MINIMUM
from ultron.sentry.api import MAXIMUM

from ultron.sentry.api import LAST
from ultron.sentry.api import CURRENT

from ultron.sentry.api import AVG
from ultron.sentry.api import DIFF
from ultron.sentry.api import SIGN
from ultron.sentry.api import EXP
from ultron.sentry.api import LOG
from ultron.sentry.api import SQRT
from ultron.sentry.api import ABS
from ultron.sentry.api import ACOS
from ultron.sentry.api import ACOSH
from ultron.sentry.api import ASIN
from ultron.sentry.api import ASINH
from ultron.sentry.api import NORMINV
from ultron.sentry.api import CEIL
from ultron.sentry.api import FLOOR
from ultron.sentry.api import ROUND
from ultron.sentry.api import SIGMOID
from ultron.sentry.api import TANH
from ultron.sentry.api import RETURNSimple
from ultron.sentry.api import RETURNLog

from ultron.sentry.api import EMA
from ultron.sentry.api import MA
from ultron.sentry.api import MADecay
from ultron.sentry.api import MMAX
from ultron.sentry.api import MARGMAX
from ultron.sentry.api import MMIN
from ultron.sentry.api import MARGMIN
from ultron.sentry.api import MRANK
from ultron.sentry.api import MQUANTILE
from ultron.sentry.api import MALLTRUE
from ultron.sentry.api import MANYTRUE
from ultron.sentry.api import MSUM
from ultron.sentry.api import MVARIANCE
from ultron.sentry.api import MSTD
from ultron.sentry.api import MNPOSITIVE
from ultron.sentry.api import MAPOSITIVE
from ultron.sentry.api import RSI
from ultron.sentry.api import MARETURNLog

from ultron.sentry.api import DELTA
from ultron.sentry.api import SHIFT

from ultron.sentry.api import MCORR
from ultron.sentry.api import MRes
from ultron.sentry.api import MSharp
from ultron.sentry.api import MSortino

import six
from enum import Enum, unique


@unique
class FunctionType(Enum):
    cross_section = 1
    time_series = 2


class Function(object):

    def __init__(self, function, arity, ftype, default_value=0):
        self.function = function
        self.arity = arity
        self.name = function.__name__
        self.ftype = ftype
        self.default_value = default_value


@six.add_metaclass(Singleton)
class Operators(object):

    def __init__(self):

        # 时间序列默认周期列表
        self._ts_period = [i for i in range(1, 22) if (i % 2) == 0]

        self._init_cs()
        self._init_ts()

        self._cs_mutation_function_list = [
            Function(f, 1, FunctionType.cross_section)
            for f in self._cross_section_mutation_list
        ]
        self._cs_crossover_function_list = [
            Function(f, 2, FunctionType.cross_section)
            for f in self._cross_section_crossover_list
        ]
        #支持全部参数
        self._ts_mutation_function_list = []
        self._ts_crossover_function_list = []
        for period in self._ts_period:
            self._ts_mutation_function_list += [
                Function(f, 1, FunctionType.time_series, period)
                for f in self._time_series_mutation_list
            ]
            self._ts_crossover_function_list += [
                Function(f, 2, FunctionType.time_series, period)
                for f in self._time_series_crossover_list
            ]

        self._mutation_sets = self._cs_mutation_function_list + self._ts_mutation_function_list
        self._crossover_sets = self._cs_crossover_function_list + self._ts_crossover_function_list

    def _init_cs(self):
        self._cross_section_mutation_list = [
            AVG, DIFF, TANH, SIGN, EXP, LOG, SQRT, ABS, ACOS, ASIN, ASINH,
            ACOSH, NORMINV, CEIL, FLOOR, ROUND, RETURNSimple, RETURNLog,
            SIGMOID, CSRank, CSZScore, CSQuantiles, CSMean, CSMeanAdjusted
        ]

        self._cross_section_crossover_list = [
            CSRes, ADDED, SUBBED, MUL, DIV, MINIMUM, MAXIMUM
        ]

    def _init_ts(self):
        self._time_series_mutation_list = [
            EMA, MA, MADecay, MMAX, MARGMAX, MMIN, MARGMIN, MRANK, MQUANTILE,
            MALLTRUE, MANYTRUE, MSUM, MVARIANCE, MSTD, MNPOSITIVE, MAPOSITIVE,
            RSI, MARETURNLog, DELTA, SHIFT
        ]

        self._time_series_crossover_list = [MCORR, MRes, MSharp, MSortino]

    def custom_transformer(self, formula_sets):
        operators_sets = self._crossover_sets + self._mutation_sets

        return [
            operator for operator in operators_sets
            if operator.name in formula_sets
        ]

    def calc_factor(self,
                    expression,
                    total_data,
                    indexs,
                    key,
                    name='transformed',
                    dropna=False):
        return eval(expression).transform(total_data.set_index(indexs),
                                          category_field=key,
                                          name=name,
                                          dropna=dropna)

    def expression_params(self, expression):
        return eval(expression)._dependency

    def fetch_mutation_sets(self):
        return self._mutation_sets

    def fetch_crossover_sets(self):
        return self._crossover_sets


mutation_sets = Operators().fetch_mutation_sets()
crossover_sets = Operators().fetch_crossover_sets()
operators_sets = mutation_sets + crossover_sets
calc_factor = Operators().calc_factor
custom_transformer = Operators().custom_transformer
expression_params = Operators().expression_params



# -*- encoding:utf-8 -*-
"""
    时间日期工具模块
"""

import datetime, time
from datetime import datetime as dt
from ultron.ump.core.fixes import six
"""默认的时间日期格式，项目中金融时间序列等时间相关默认格式"""
K_DEFAULT_DT_FMT = "%Y-%m-%d"


def str_to_datetime(date_str, fmt=K_DEFAULT_DT_FMT, fix=True):
    """
    将字符串日期格式转换成datetime.datetime对象 eg. '2016-01-01' －> datetime.datetime(2016, 1, 1, 0, 0)
    :param date_str: %Y-%m-%d 形式str对象，eg. '2016-01-01'
    :param fmt: 如date_str不是%Y-%m-%d形式，对应的格式str对象
    :param fix: 是否修复日期不规范的写法，eg. 2016-1-1 fix 2016-01-01
    :return: datetime.datetime对象，eg. datetime.datetime(2016, 1, 1, 0, 0)
    """
    if fix and fmt == K_DEFAULT_DT_FMT:
        # 只针对%Y-%m-%d形式格式标准化日期格式
        date_str = fix_date(date_str)

    return dt.strptime(date_str, fmt)


def timestamp_to_str(ts):
    """
    针对pandas.tslib.Timestamp对象，即金融时间序列index元素对象转换为str对象, 时间单位只取到天，返回如2016-01－01
    :param ts: pandas.tslib.Timestamp对象，eg. Timestamp('2016-01-01 00:00:00')
    :return: 回如2016-01－01 str对象
    """
    try:
        # pandas高版本to_pydatetime
        s_date = str(ts.to_pydatetime().date())[:10]
    except:
        # pandas低版本to_datetime
        s_date = str(ts.to_datetime().date())[:10]
    return s_date


def date_str_to_int(date_str, split='-', fix=True):
    """
    eg. 2016-01-01 -> 20160101
    不使用时间api，直接进行字符串解析，执行效率高
    :param date_str: %Y-%m-%d形式时间str对象
    :param split: 年月日的分割符，默认'-'
    :param fix: 是否修复日期不规范的写法，eg. 2016-1-1 fix 2016-01-01
    :return: int类型时间
    """
    if fix and split == '-':
        # 只针对%Y-%m-%d形式格式标准化日期格式
        date_str = fix_date(date_str)
    string_date = date_str.replace(split, '')
    return int(string_date)


def week_of_date(date_str, fmt=K_DEFAULT_DT_FMT, fix=True):
    """
    输入'2016-01-01' 转换为星期几，返回int 0-6分别代表周一到周日
    :param date_str: 式时间日期str对象
    :param fmt: 如date_str不是%Y-%m-%d形式，对应的格式str对象
    :param fix: 是否修复日期不规范的写法，eg. 2016-1-1 fix 2016-01-01
    :return: 返回int 0-6分别代表周一到周日
    """

    if fix and fmt == K_DEFAULT_DT_FMT:
        # 只针对%Y-%m-%d形式格式标准化日期格式
        date_str = fix_date(date_str)
    return dt.strptime(date_str, fmt).weekday()


def fix_date(date_str):
    """
    修复日期不规范的写法:
                eg. 2016-1-1 fix 2016-01-01
                eg. 2016:01-01 fix 2016-01-01
                eg. 2016,01 01 fix 2016-01-01
                eg. 2016/01-01 fix 2016-01-01
                eg. 2016/01/01 fix 2016-01-01
                eg. 2016/1/1 fix 2016-01-01
                eg. 2016:1:1 fix 2016-01-01
                eg. 2016 1 1 fix 2016-01-01
                eg. 2016 01 01 fix 2016-01-01
                .............................
    不使用时间api，直接进行字符串解析，执行效率高，注意fix_date内部会使用fmt_date
    :param date_str: 检测需要修复的日期str对象或者int对象
    :return: 修复了的日期str对象
    """
    if date_str is not None:
        # 如果是字符串先统一把除了数字之外的都干掉，变成干净的数字串
        if isinstance(date_str, six.string_types):
            # eg, 2016:01-01, 201601-01, 2016,01 01, 2016/01-01 -> 20160101
            date_str = ''.join(list(filter(lambda c: c.isdigit(), date_str)))
        # 再统一确定%Y-%m-%d形式
        date_str = fmt_date(date_str)
        y, m, d = date_str.split('-')
        if len(m) == 1:
            # 月上补0
            m = '0{}'.format(m)
        if len(d) == 1:
            # 日上补0
            d = '0{}'.format(d)
        date_str = "%s-%s-%s" % (y, m, d)
    return date_str


def str_to_datetime_fast(date_str, split='-', fix=True):
    """
    不使用datetime api直接进行字符串分解，相对str_to_datetime要快很多，大概2倍的时间，大量时间转换，且格式确定使用
    将字符串日期格式转换成datetime.datetime对象 eg. '2016-01-01' －> datetime.datetime(2016, 1, 1, 0, 0)
    :param date_str: 如date_str不是%Y-%m-%d形式，对应的格式str对象
    :param split: 年月日的分割符，默认'-'
    :param fix: 是否修复日期不规范的写法，eg. 2016-1-1 fix 2016-01-01
    :return: datetime.datetime对象，eg. datetime.datetime(2016, 1, 1, 0, 0)
    """
    if fix and split == '-':
        # 只针对%Y-%m-%d形式格式标准化日期格式
        date_str = fix_date(date_str)
    y, m, d = date_str.split(split)
    return dt(int(y), int(m), int(d))


def diff(start_date, end_date, check_order=True):
    """
    对两个输入日期计算间隔的天数，如果check_order=False, str日期对象效率最高
    :param start_date: str对象或者int对象，如果check_order=True int对象效率最高
    :param end_date: str对象或者int对象，如果check_order=True int对象效率最高
    :param check_order: 是否纠正参数顺序是否放置正常，默认check
    :return:
    """

    # 首先进来的date都格式化，主要进行的是fix操作，不管是int还是str，这样20160101转换为2016-01-01日期格式
    start_date = fix_date(start_date)
    end_date = fix_date(end_date)

    if check_order and isinstance(start_date, six.string_types):
        # start_date字符串的日期格式转换为int
        start_date = date_str_to_int(start_date)

    if check_order and isinstance(end_date, six.string_types):
        # end_date字符串的日期格式转换为int
        end_date = date_str_to_int(end_date)

    # 是否纠正参数顺序是否放置正常，默认check
    if check_order and start_date > end_date:
        # start_date > end_date说明要换一下
        tmp = end_date
        end_date = start_date
        start_date = tmp

    # fmt_date，但在不需要纠正check_order的情况，这些就都不会执行
    if isinstance(start_date, int):
        # noinspection PyTypeChecker
        start_date = fmt_date(start_date)
    if isinstance(end_date, int):
        # noinspection PyTypeChecker
        end_date = fmt_date(end_date)

    # 在不需要纠正check_order的情况, 直接执行的是这里
    sd = str_to_datetime(start_date)
    ed = str_to_datetime(end_date)

    return (ed - sd).days


def fmt_date(convert_date):
    """
    将时间格式如20160101转换为2016-01-01日期格式, 注意没有对如 201611
    这样的做fix适配，外部需要明确知道参数的格式，针对特定格式，不使用时间api，
    直接进行字符串解析，执行效率高
    :param convert_date: 时间格式如20160101所示，int类型或者str类型对象
    :return: %Y-%m-%d日期格式str类型对象
    """
    if isinstance(convert_date, float):
        # float先转换int
        convert_date = int(convert_date)
    convert_date = str(convert_date)

    if len(convert_date) > 8 and convert_date.startswith('20'):
        # eg '20160310000000000'
        convert_date = convert_date[:8]

    if '-' not in convert_date:
        if len(convert_date) == 8:
            # 20160101 to 2016-01-01
            convert_date = "%s-%s-%s" % (convert_date[0:4], convert_date[4:6],
                                         convert_date[6:8])
        elif len(convert_date) == 6:
            # 201611 to 2016-01-01
            convert_date = "%s-0%s-0%s" % (
                convert_date[0:4], convert_date[4:5], convert_date[5:6])
        else:
            raise ValueError(
                'fmt_date: convert_date fmt error {}'.format(convert_date))
    return convert_date


def current_date_int():
    """
    获取当前时间日期 int值
    不使用时间api，直接进行字符串解析，执行效率高
    :return: 日期int值
    """
    date_int = 0
    # 首先获取str
    today = current_str_date()
    # 手动比系统a时间pi快
    today_array = today.split("-")
    if len(today_array) == 3:
        date_int = int(today_array[0]) * 10000 + int(
            today_array[1]) * 100 + int(today_array[2])
    return date_int


def current_str_date():
    """
    获取当前时间日期，时间单位只取到天，返回如2016-01－01
    :return: 返回如2016-01－01 str对象
    """
    return str(datetime.date.today())


# -*- encoding:utf-8 -*-
"""
    基准模块，基准的作用在于交易时间范围确定，交易时间序列对齐，
    抛弃异常时间序列，交易市场范围限制，以及对比与策略的度量结果等作用
"""
from ultron.ump.core.base import PickleStateMixin
from ultron.ump.core.fixes import six
from ultron.ump.market.symbol import Symbol


class Benchmark(PickleStateMixin):
    """基准类，混入PickleStateMixin，因为在abu.store_abu_result_tuple会进行对象本地序列化"""

    def __init__(self,
                 benchmark=None,
                 start=None,
                 end=None,
                 n_folds=2,
                 rs=True,
                 benchmark_kl_pd=None):
        if benchmark_kl_pd is not None and hasattr(benchmark_kl_pd, 'name'):
            """从金融时间序列直接构建"""
            self.benchmark = benchmark_kl_pd.name
            self.start = benchmark_kl_pd.iloc[0].date
            self.end = benchmark_kl_pd.iloc[-1].date
            self.n_folds = n_folds
            self.kl_pd = benchmark_kl_pd
            return

        self.benchmark = benchmark
        self.start = start
        self.end = end
        self.n_folds = n_folds
        # 基准获取数据使用data_mode=EMarketDataSplitMode.E_DATA_SPLIT_SE，即不需要对齐其它，只需要按照时间切割
        self.kl_pd = None  ## fix me
        #self.kl_pd = ABuSymbolPd.make_kl_df(
        #    benchmark,
        #    data_mode=EMarketDataSplitMode.E_DATA_SPLIT_SE,
        #    n_folds=n_folds,
        #    start=start,
        #    end=end)

        if rs and self.kl_pd is None:
            # 如果基准时间序列都是none，就不要再向下运行了
            raise ValueError('CapitalClass init benchmark kl_pd is None')

    def unpick_extend_work(self, state):
        """完成 PickleStateMixin中__setstate__结束之前的工作，为kl_pd.name赋予准确的benchmark"""
        if isinstance(self.benchmark, Symbol):
            self.kl_pd.name = self.benchmark.value
        elif isinstance(self.benchmark, six.string_types):
            self.kl_pd.name = self.benchmark

    def __str__(self):
        """打印对象显示：benchmark n_folds"""
        return 'benchmark is {}, n_folds = {}'.format(self.kl_pd.name,
                                                      self.n_folds)

    __repr__ = __str__


# -*- encoding:utf-8 -*-
"""
    内置特征定义，以及用户特征扩展，定义模块
"""
import numpy as np
import ast
from ultron.kdutils import regression
from ultron.ump.core import env
from ultron.ump.core.fixes import xrange, six
from ultron.ump.technical.wave import calc_wave_std
from ultron.ump.technical.atr import calc_atr_std
from ultron.ump.technical.jump import calc_jump

g_market_trade_year = 252
# 内置特征，趋势角度
g_deg_keys = [21, 42, 60, g_market_trade_year]
# 内置特征，价格rank
g_price_rank_keys = [60, 90, 120, g_market_trade_year]
# 内置特征，波动周期定义
g_wave_xd = 42
# 内置特征，波动取样个数
g_wave_key_cnt = 3
# 内置特征，atr周期定义
g_atr_xd = 42


class BuyFeatureMixin(object):
    """
        买入特征标识混入，与BuyUmpMixin不同，具体feature类可能属于多个类别
        即可能同时混入BuyFeatureMixin和SellFeatureMixin
    """
    _feature_buy = True
    _feature_buy_prefix = 'buy_'


class SellFeatureMixin(object):
    """
        卖出特征标识混入，与SellUmpMixin不同，具体feature类可能属于多个类别
        即可能同时混入BuyFeatureMixin和SellFeatureMixin
    """
    _feature_sell = True
    _feature_sell_prefix = 'sell_'


class FeatureBase(object):
    """特征构造基类"""

    def support_buy_feature(self):
        """是否支持买入特征构建"""
        return getattr(self, "_feature_buy", False) is True

    def support_sell_feature(self):
        """是否支持卖出特征构建"""
        return getattr(self, "_feature_sell", False) is True

    def check_support(self, buy_feature):
        """
        根据参数buy_feature检测是否支持特征构建
        :param buy_feature: 是否是买入特征构造（bool）
        """
        if buy_feature and not self.support_buy_feature:
            raise TypeError(
                'feature support buy must subclass BuyFeatureMixin!!!')
        if not buy_feature and not self.support_sell_feature:
            raise TypeError(
                'feature support buy must subclass SellFeatureMixin!!!')

    def feature_prefix(self, buy_feature, check=True):
        """
        根据buy_feature决定返回_feature_buy_prefix或者_feature_sell_prefix，目的是在calc_feature中构成唯一key
        :param buy_feature: 是否是买入特征构造（bool）
        :param check: 是否需要检测是否支持特征构建
        :return:
        """
        if check:
            self.check_support(buy_feature)

        return getattr(self,
                       '_feature_buy_prefix') if buy_feature else getattr(
                           self, '_feature_sell_prefix')

    def __str__(self):
        """打印对象显示：class name, support_buy_feature support_sell_feature, get_feature_keys"""
        return '{}:is_buy_feature:{} is_sell_feature:{} feature: {}'.format(
            self.__class__.__name__, self.support_buy_feature(),
            self.support_sell_feature(),
            self.get_feature_keys(self.support_buy_feature()))

    __repr__ = __str__

    def get_feature_ump_keys(self, ump_cls):
        """
        根据ump_cls，返回对应的get_feature_keys
        :param ump_cls: UltronUmpEdgeBase子类，参数为类，非实例对象
        :return: 键值对字典中的key序列
        """
        is_buy_ump = getattr(ump_cls, "_ump_type_prefix") == 'buy_'
        return self.get_feature_keys(buy_feature=is_buy_ump)

    def get_feature_keys(self, buy_feature):
        """
        子类主要需要实现的函数，定义feature的列名称
        :param buy_feature: 是否是买入特征构造（bool）
        :return: 键值对字典中的key序列
        """
        raise NotImplementedError('NotImplementedError get_feature_keys!!!')

    def calc_feature(self, kl_pd, combine_kl_pd, day_ind, buy_feature):
        """
        子类主要需要实现的函数，根据买入或者卖出时的金融时间序列，以及交易日信息构造特征
        :param kl_pd: 择时阶段金融时间序列
        :param combine_kl_pd: 合并择时阶段之前1年的金融时间序列
        :param day_ind: 交易发生的时间索引，即对应self.kl_pd.key
        :param buy_feature: 是否是买入特征构造（bool）
        :return: 构造特征的键值对字典
        """
        raise NotImplementedError('NotImplementedError calc_feature!!!')


class FeatureDeg(FeatureBase, BuyFeatureMixin, SellFeatureMixin):
    """角度特征，支持买入，卖出"""

    def __init__(self):
        """
            默认21, 42, 60, 250日走势角度特征，如外部修改，直接使用类似如下：
                       ultronpy.feature.g_deg_keys = [10, 20, 30, 40, 50]
        """
        # frozenset包一下，一旦定下来就不能修改，否则特征对不上
        self.deg_keys = frozenset(g_deg_keys)

    def get_feature_keys(self, buy_feature):
        """
        迭代生成所有走势角度特征feature的列名称定, 使用feature_prefix区分买入，卖出前缀key
        :param buy_feature: 是否是买入特征构造（bool）
        :return: 角度特征的键值对字典中的key序列
        """
        return [
            '{}deg_ang{}'.format(self.feature_prefix(buy_feature=buy_feature),
                                 dk) for dk in self.deg_keys
        ]

    def calc_feature(self, kl_pd, combine_kl_pd, day_ind, buy_feature):
        """
        根据买入或者卖出时的金融时间序列，以及交易日信息构造拟合角度特征
        :param kl_pd: 择时阶段金融时间序列
        :param combine_kl_pd: 合并择时阶段之前1年的金融时间序列
        :param day_ind: 交易发生的时间索引，即对应self.kl_pd.key
        :param buy_feature: 是否是买入特征构造（bool）
        :return: 构造角度特征的键值对字典
        """
        # 返回的角度特征键值对字典
        deg_dict = {}
        for dk in self.deg_keys:
            # 迭代预设角度周期，计算构建特征
            if day_ind - dk >= 0:
                # 如果择时时间序列够提取特征，使用kl_pd截取特征交易周期收盘价格
                deg_close = kl_pd[day_ind - dk + 1:day_ind + 1].close
            else:
                # 如果择时时间序列不够提取特征，使用combine_kl_pd截取特征交易周期，首先截取直到day_ind的时间序列
                combine_kl_pd = combine_kl_pd.loc[:kl_pd.index[day_ind]]
                # 如combine_kl_pd长度大于特征周期长度－> 截取combine_kl_pd[-dk:].close，否则取combine_kl_pd所有交易收盘价格
                deg_close = combine_kl_pd[-dk:].close if combine_kl_pd.shape[
                    0] > dk else combine_kl_pd.close
            # 使用截取特征交易周期收盘价格deg_close做为参数，通过calc_regress_deg计算趋势拟合角度
            ang = regression.calc_regress_deg(deg_close)
            # 标准化拟合角度值
            ang = 0 if np.isnan(ang) else round(ang, 3)
            # 角度特征键值对字典添加拟合角度周期key和对应的拟合角度值
            deg_dict['{}deg_ang{}'.format(
                self.feature_prefix(buy_feature=buy_feature), dk)] = ang

        return deg_dict


class FeaturePrice(FeatureBase, BuyFeatureMixin, SellFeatureMixin):
    """价格rank特征，支持买入，卖出"""

    def __init__(self):
        """
            默认60, 90, 120, 250日价格rank特征，如外部修改，直接使用类似如下：
                    ultronpy.feature.g_price_rank_keys = [10, 20, 30, 40，50]
        """
        # frozenset包一下，一旦定下来就不能修改，否则特征对不上
        self.price_rank_keys = frozenset(g_price_rank_keys)

    def get_feature_keys(self, buy_feature):
        """
        迭代生成所有价格rank特征feature的列名称, 使用feature_prefix区分买入，卖出前缀key
        :param buy_feature: 是否是买入特征构造（bool）
        :return: 价格rank特征的键值字典中的key序列
        """
        return [
            '{}price_rank{}'.format(
                self.feature_prefix(buy_feature=buy_feature), dk)
            for dk in self.price_rank_keys
        ]

    def calc_feature(self, kl_pd, combine_kl_pd, day_ind, buy_feature):
        """
        根据买入或者卖出时的金融时间序列，以及交易日信息构造波动特征
        :param kl_pd: 择时阶段金融时间序列
        :param combine_kl_pd: 合并择时阶段之前1年的金融时间序列
        :param day_ind: 交易发生的时间索引，即对应self.kl_pd.key
        :param buy_feature: 是否是买入特征构造（bool）
        :return: 构造波动特征的键值对字典
        """
        # 返回的波动特征键值对字典
        price_rank_dict = {}
        for dk in self.price_rank_keys:
            # 迭代预设价格rank周期，计算构建特征
            if day_ind - dk >= 0:
                # 如果择时时间序列够提取特征，使用kl_pd截取特征交易周期收盘价格
                price_close = kl_pd[day_ind - dk + 1:day_ind + 1].close
            else:
                # 如果择时时间序列不够提取特征，使用combine_kl_pd截取特征交易周期，首先截取直到day_ind的时间序列
                combine_kl_pd = combine_kl_pd.loc[:kl_pd.index[day_ind]]
                # 如combine_kl_pd长度大于特征周期长度－> 截取combine_kl_pd[-dk:].close，否则取combine_kl_pd所有交易收盘价格
                price_close = combine_kl_pd[-dk:].close if combine_kl_pd.shape[
                    0] > dk else combine_kl_pd.close
            """
                price_close.rank()的结果是所有价格的排名

                ...................
                2016-07-20    256.0
                2016-07-21    200.0
                2016-07-22    214.0
                2016-07-25    266.0
                2016-07-26    239.0

                -> price_close.rank()[-1]的结果是买入时刻或者卖出时刻的排名

                239.0

                －> price_close.rank()[-1] / price_close.rank().shape[0] 的结果即为：
                买入时刻或者卖出时刻的排名在周期中的排名位置，值由0-1

                eg: price_close.rank()[-1] / price_close.rank().shape[0]
                -> 239.0 / 504 = 0.47420634920634919, 即代表买入或者卖出时价格在特征周期中的位置
            """
            price_rank = price_close.rank()[-1] / price_close.rank().shape[0]
            # 标准化价格rank值
            price_rank = 0 if np.isnan(price_rank) else round(price_rank, 3)
            # 价格rank特征键值对字典添加价格rank周期key和对应的价格rank值
            price_rank_dict['{}price_rank{}'.format(
                self.feature_prefix(buy_feature=buy_feature), dk)] = price_rank
        return price_rank_dict


class FeatureWave(FeatureBase, BuyFeatureMixin, SellFeatureMixin):
    """波动特征，支持买入，卖出"""

    def __init__(self):
        """
            默认42日做为波动计算周期，如外部修改，直接使用类似如下：
                ultronpy.feature.g_wave_xd = 21
        """
        self.wave_xd = g_wave_xd
        self.wave_key_cnt = g_wave_key_cnt

    def get_feature_keys(self, buy_feature):
        """
        迭代生成所有波动特征feature的列名称, 使用feature_prefix区分买入，卖出前缀key
        :param buy_feature: 是否是买入特征构造（bool）
        :return: 波动特征的键值对字典中的key序列
        """
        return [
            '{}wave_score{}'.format(
                self.feature_prefix(buy_feature=buy_feature), xd_ind)
            for xd_ind in list(range(1, self.wave_key_cnt + 1))
        ]

    def calc_feature(self, kl_pd, combine_kl_pd, day_ind, buy_feature):
        """
        根据买入或者卖出时的金融时间序列，以及交易日信息构造波动特征
        :param kl_pd: 择时阶段金融时间序列
        :param combine_kl_pd: 合并择时阶段之前1年的金融时间序列
        :param day_ind: 交易发生的时间索引，即对应self.kl_pd.key
        :param buy_feature: 是否是买入特征构造（bool）
        :return: 构造波动特征的键值对字典
        """
        # 构建波动特征周期固定，即一年的交易日数量
        wave_wide = g_market_trade_year
        if day_ind - wave_wide >= 0:
            # 如果择时时间序列够提取特征，使用kl_pd截取特征交易周期收盘价格
            wave_df = kl_pd[day_ind - wave_wide + 1:day_ind + 1]
        else:
            # 如果择时时间序列不够提取特征，使用combine_kl_pd截取特征交易周期，首先截取直到day_ind的时间序列
            combine_kl_pd = combine_kl_pd.loc[:kl_pd.index[day_ind]]
            # 如combine_kl_pd长度大于特征周期长度－> 截取combine_kl_pd[-wave_wide:].close，否则取combine_kl_pd所有交易收盘价格
            wave_df = combine_kl_pd[-wave_wide:] if combine_kl_pd.shape[
                0] > wave_wide else combine_kl_pd
        # 返回的波动特征键值对字典
        wave_dict = {}
        for xd_ind in xrange(1, self.wave_key_cnt + 1):
            # wave_df固定为一年交易时间序列，xd是calc_wave_st内部计算rolling std的window值，详UltronTLWave.calc_wave_std
            wave = calc_wave_std(wave_df, xd=xd_ind * self.wave_xd)
            wave_score = wave.score
            # 标准化波动特征值
            wave_score = 0 if np.isnan(wave_score) else round(wave_score, 3)
            # 波动特征键值对字典添加波动特征key和对应的波动特征值
            wave_dict['{}wave_score{}'.format(
                self.feature_prefix(buy_feature=buy_feature),
                xd_ind)] = wave_score
        return wave_dict


class FeatureAtr(FeatureBase, BuyFeatureMixin):
    """atr特征，支持买入"""

    def __init__(self):
        """
            默认42日做为atr特征计算周期，如外部修改，直接使用类似如下：
                ultronpy.feature.g_atr_xd = 21
        """
        self.atr_xd = g_atr_xd
        self.atr_key = 'atr_std'

    def get_feature_keys(self, buy_feature):
        """
        返回对应的atr特征key值，虽然只有一个固定的atr key值，也返回序列，保持接口统一
        :param buy_feature: 是否是买入特征构造（bool）
        :return: atr特征的键值对字典中的key序列
        """
        return [
            '{}{}'.format(self.feature_prefix(buy_feature=buy_feature),
                          self.atr_key)
        ]

    def calc_feature(self, kl_pd, combine_kl_pd, day_ind, buy_feature):
        """
        根据买入或者卖出时的金融时间序列，以及交易日信息构造atr特征
        :param kl_pd: 择时阶段金融时间序列
        :param combine_kl_pd: 合并择时阶段之前1年的金融时间序列
        :param day_ind: 交易发生的时间索引，即对应self.kl_pd.key
        :param buy_feature: 是否是买入特征构造（bool）
        :return: 构造atr特征的键值对字典
        """
        # 构建atr特征周期固定，即一年的交易日数量
        atr_wide = g_market_trade_year
        if day_ind - atr_wide >= 0:
            # 如果择时时间序列够提取特征，使用kl_pd截取特征交易周期收盘价格
            atr_df = kl_pd[day_ind - atr_wide + 1:day_ind + 1]
        else:
            # 如果择时时间序列不够提取特征，使用combine_kl_pd截取特征交易周期，首先截取直到day_ind的时间序列
            combine_kl_pd = combine_kl_pd.loc[:kl_pd.index[day_ind]]
            # 如combine_kl_pd长度大于特征周期长度－> 截取combine_kl_pd[-atr_wide:].close，否则取combine_kl_pd所有交易收盘价格
            atr_df = combine_kl_pd[-atr_wide:] if combine_kl_pd.shape[
                0] > atr_wide else combine_kl_pd
        # 返回的atr特征键值对字典
        atr_dict = {}
        # 计算atr特征，详见UltronTLAtr.calc_atr_std
        atr_std = calc_atr_std(atr_df, xd=self.atr_xd)
        atr_score = atr_std.score
        # 标准化atr特征
        atr_score = 0 if np.isnan(atr_score) else round(atr_score, 3)
        # atr特征键值对字典添加atr特征key和对应的atr特征值
        atr_dict['{}{}'.format(self.feature_prefix(buy_feature=buy_feature),
                               self.atr_key)] = atr_score
        return atr_dict


class FeatureJump(FeatureBase, BuyFeatureMixin, SellFeatureMixin):
    """跳空特征，支持买入，卖出"""

    def __init__(self):
        """
            跳空的特征key为:
            jump_down_power: 向下跳空的能量
            diff_down_days : 向下跳空距离买入或卖出的时间间隔
            jump_up_power  : 向上跳空的能量
            diff_up_days   : 向上跳空距离买入或卖出的时间间隔
        """
        self.jump_keys = frozenset([
            'jump_down_power', 'diff_down_days', 'jump_up_power',
            'diff_up_days'
        ])

    def get_feature_keys(self, buy_feature):
        """
        返回对应的跳空特征key值
        :param buy_feature: 是否是买入特征构造（bool）
        :return: 跳空特征的键值对字典中的key序列
        """
        return [
            '{}{}'.format(self.feature_prefix(buy_feature=buy_feature), jk)
            for jk in self.jump_keys
        ]

    def calc_feature(self, kl_pd, combine_kl_pd, day_ind, buy_feature):
        """
        根据买入或者卖出时的金融时间序列，以及交易日信息构造跳空特征
        :param kl_pd: 择时阶段金融时间序列
        :param combine_kl_pd: 合并择时阶段之前1年的金融时间序列
        :param day_ind: 交易发生的时间索引，即对应self.kl_pd.key
        :param buy_feature: 是否是买入特征构造（bool）
        :return: 构造跳空特征的键值对字典
        """

        # 构建跳空特征周期固定，即一年的交易日数量
        jump_wide = g_market_trade_year
        if day_ind - jump_wide >= 0:
            # 如果择时时间序列够提取特征，使用kl_pd截取特征交易周期收盘价格
            jump_df = kl_pd[day_ind - jump_wide + 1:day_ind + 1]
        else:
            # 如果择时时间序列不够提取特征，使用combine_kl_pd截取特征交易周期，首先截取直到day_ind的时间序列
            combine_kl_pd = combine_kl_pd.loc[:kl_pd.index[day_ind]]
            # 如combine_kl_pd长度大于特征周期长度－> 截取combine_kl_pd[-jump_wide:].close，否则取combine_kl_pd所有交易收盘价格
            jump_df = combine_kl_pd[-jump_wide:] if combine_kl_pd.shape[
                0] > jump_wide else combine_kl_pd

        # 返回的跳空特征键值对字典
        jump_dict = {}

        diff_down_days = 0
        jump_down_power = 0
        diff_up_days = 0
        jump_up_power = 0

        key_prefix = self.feature_prefix(buy_feature=buy_feature)
        # 通过特征周期时间序列jump_df做为参数，计算跳空，返回jumps对象为pd.DataFrame对象，详见UltronTLJump.calc_jump
        jumps = calc_jump(jump_df)
        """
                jump形式如下所示：jump代表跳空方向1为向上，－1为向下，jump_power代表对应的跳空能量，即每一行
                记录了一次跳空发生的时间，价格变化，跳空能量等信息
                            jump	jump_power	close	date	p_change	pre_close
                2014-08-11	 1.0	1.006085	259.32	20140811.0	4.51	248.13
                2014-10-10	-1.0	1.628481	236.91	20141010.0	-7.82	257.01
                2015-01-14	-1.0	1.325337	192.69	20150114.0	-5.66	204.25
                2015-02-12	-1.0	1.422285	202.88	20150212.0	-4.66	212.80
                .............
        """
        if not jumps.empty:
            # 筛选出所有向下跳空的
            down_jumps = jumps[(jumps.jump == -1)]
            if down_jumps.shape[0] > 0:
                # 筛选最后一次向下跳空的情况
                last_down_jump = down_jumps.iloc[-1:]
                # 跳空能力 * last_down_jump.jump.values[0] 转换成有方向的
                jump_down_power = last_down_jump.jump_power.values[
                    0] * last_down_jump.jump.values[0]
                # 向下跳空距离买入或卖出的时间间隔
                diff_down_days = (kl_pd.iloc[day_ind:day_ind + 1].index.date -
                                  last_down_jump.index.date)[0].days
            # 筛选出所有向上跳空的
            up_jumps = jumps[(jumps.jump == 1)]
            if up_jumps.shape[0] > 0:
                # 筛选最后一次向上跳空的情况
                last_up_jump = up_jumps.iloc[-1:]
                # 跳空能力* last_up_jump.jump.values[0] 转换成有方向的
                jump_up_power = last_up_jump.jump_power.values[
                    0] * last_up_jump.jump.values[0]
                # 向上跳空距离买入或卖出的时间间隔
                diff_up_days = (kl_pd.iloc[day_ind:day_ind + 1].index.date -
                                last_up_jump.index.date)[0].days

        # 标准化跳空特征特征值
        jump_down_power = round(jump_down_power, 3)
        jump_up_power = round(jump_up_power, 3)

        # 跳空特征键值对字典添加跳空特征key和对应的跳空特征值
        jump_dict['{}jump_down_power'.format(key_prefix)] = jump_down_power
        jump_dict['{}diff_down_days'.format(key_prefix)] = diff_down_days

        jump_dict['{}jump_up_power'.format(key_prefix)] = jump_up_power
        jump_dict['{}diff_up_days'.format(key_prefix)] = diff_up_days

        return jump_dict


"""用户可扩展自定义特征"""
_g_extend_feature_list = list()


def append_user_feature(feature, check=True):
    """
    外部设置扩展feature接口
    :param feature: 可以是feature class类型，也可以是实例化后的feature object
    :param check: 是否检测feature是UltronFeatureBase实例
    :return:
    """
    if isinstance(feature, six.class_types):
        # 暂时认为所有feature的实例化不需要参数，如需要也可添加＊args
        feature_obj = feature()
    else:
        # 如果不是类直接赋值
        feature_obj = feature

    # check检测feature_obj是不是FeatureBase的子类实例对象
    if check and not isinstance(feature_obj, FeatureBase):
        raise TypeError('feature must a isinstance UltronFeatureBase!!!')

    # 添加到用户可扩展自定义特征序列中
    _g_extend_feature_list.append(feature_obj)


def clear_user_feature():
    """将用户可扩展自定义特征序列清空"""
    global _g_extend_feature_list
    _g_extend_feature_list = list()


class MlFeature(object):
    """特征对外统一接口类，负责管理构建内部特征，用户扩展特征，提供买入卖出因子的交易特征生成，转换接口"""

    def __init__(self):
        """实例化 内置特征对象＋用户扩展自定义特征对象"""
        self.features = [
            FeatureDeg(),
            FeaturePrice(),
            FeatureWave(),
            FeatureAtr(),
            FeatureJump()
        ]
        # 用户扩展自定义特征对象extend到特征对象序列self.features中
        if len(_g_extend_feature_list) > 0:
            # 这里不再次check了，因为append_user_feature时已经做过，直接extend
            self.features.extend(_g_extend_feature_list)

    def make_feature_dict(self, kl_pd, combine_kl_pd, day_ind, buy_feature):
        """
        提供买入卖出因子构建交易特征的接口，使用见UltronFactorBuyBase.make_buy_order_ml_feature
        :param kl_pd: 择时阶段金融时间序列
        :param combine_kl_pd: 合并择时阶段之前1年的金融时间序列
        :param day_ind: 交易发生的时间索引，即对应self.kl_pd.key
        :param buy_feature: 是否是买入特征构造（bool）
        """
        if not env.g_enable_ml_feature:
            # 需要env中开启特征生成，否则不生成交易特征
            return None

        ml_feature_dict = {}

        # 根据buy_feature，过滤出特征序列中支持的特征序列子集
        features = list(
            filter(
                lambda f: f.support_buy_feature()
                if buy_feature else f.support_sell_feature(), self.features))
        for feature in features:
            # 迭代特征序列对象，特征对象统一使用calc_feature接口生成自己的特征，结果特征update到总特征字典ml_feature_dict中
            ml_feature_dict.update(
                feature.calc_feature(kl_pd, combine_kl_pd, day_ind,
                                     buy_feature))
        return ml_feature_dict

    def _get_unzip_feature_keys(self, buy_feature):
        """
        根据buy_feature，过滤出特征序列中支持的特征序列子集，迭代特征序列对象，
        使用特征对象的get_feature_keys方法获取特征的key序列
        :param buy_feature: 是否是买入特征构造（bool）
        """
        feature_key_list = []
        # 根据buy_feature，过滤出特征序列中支持的特征序列子集
        features = list(
            filter(
                lambda f: f.support_buy_feature()
                if buy_feature else f.support_sell_feature(), self.features))

        # 迭代特征序列对象，使用特征对象的get_feature_keys方法获取特征的key序列
        for feature in features:
            feature_key_list.extend(feature.get_feature_keys(buy_feature))
        return feature_key_list

    def unzip_ml_feature(self, orders_pd):
        """
        UltronTradeExecute中make_orders_pd使用，将order中dict字典形式的特征解压拆解为独立的
        pd.DataFrame列，即一个特征key，对应一个列
        :param orders_pd: 回测结果生成的交易行为构成的pd.DataFrame对象
        :return:
        """
        if env.g_enable_ml_feature:
            features_keys = list()
            # 收集买入特征keys
            features_keys.extend(self._get_unzip_feature_keys(True))
            # 收集卖出特征keys
            features_keys.extend(self._get_unzip_feature_keys(False))

            # from ..UtilBu.UltronDTUtil import except_debug
            # @except_debug
            def map_str_dict(order, key):
                if order.sell_type == 'keep' and key.startswith('sell_'):
                    # 针对卖出特征值，如果单子keep状态，即没有特征值
                    return np.nan

                if not isinstance(order.ml_features, dict):
                    # 低版本pandas dict对象取出来会成为str
                    map_ast = ast.literal_eval(order.ml_features)[key]
                else:
                    map_ast = order.ml_features[key]

                return map_ast

            for fk in features_keys:
                # 迭代所有key，fk做为pd.DataFrame对象orders_pd的新列名，
                orders_pd[fk] = orders_pd.apply(map_str_dict,
                                                axis=1,
                                                args=(fk, ))


class FeatureDegExtend(FeatureBase, BuyFeatureMixin, SellFeatureMixin):
    """示例添加新的视角来录制比赛，角度特征，支持买入，卖出"""

    def __init__(self):
        """20, 40, 60, 90, 120日走势角度特征"""
        # frozenset包一下，一旦定下来就不能修改，否则特征对不上
        self.deg_keys = frozenset([10, 30, 50, 90, 120])

    def get_feature_keys(self, buy_feature):
        """
        迭代生成所有走势角度特征feature的列名称定, 使用feature_prefix区分买入，卖出前缀key
        :param buy_feature: 是否是买入特征构造（bool）
        :return: 角度特征的键值对字典中的key序列
        """
        return [
            '{}deg_ang{}'.format(self.feature_prefix(buy_feature=buy_feature),
                                 dk) for dk in self.deg_keys
        ]

    def calc_feature(self, kl_pd, combine_kl_pd, day_ind, buy_feature):
        """
        根据买入或者卖出时的金融时间序列，以及交易日信息构造拟合角度特征
        :param kl_pd: 择时阶段金融时间序列
        :param combine_kl_pd: 合并择时阶段之前1年的金融时间序列
        :param day_ind: 交易发生的时间索引，即对应self.kl_pd.key
        :param buy_feature: 是否是买入特征构造（bool）
        :return: 构造角度特征的键值对字典
        """
        # 返回的角度特征键值对字典
        deg_dict = {}
        for dk in self.deg_keys:
            # 迭代预设角度周期，计算构建特征
            if day_ind - dk >= 0:
                # 如果择时时间序列够提取特征，使用kl_pd截取特征交易周期收盘价格
                deg_close = kl_pd[day_ind - dk + 1:day_ind + 1].close
            else:
                # 如果择时时间序列不够提取特征，使用combine_kl_pd截取特征交易周期，首先截取直到day_ind的时间序列
                combine_kl_pd = combine_kl_pd.loc[:kl_pd.index[day_ind]]
                # 如combine_kl_pd长度大于特征周期长度－> 截取combine_kl_pd[-dk:].close，否则取combine_kl_pd所有交易收盘价格
                deg_close = combine_kl_pd[-dk:].close if combine_kl_pd.shape[
                    0] > dk else combine_kl_pd.close

            # 使用截取特征交易周期收盘价格deg_close做为参数，通过calc_regress_deg计算趋势拟合角度
            ang = regression.calc_regress_deg(deg_close)
            # 标准化拟合角度值
            ang = 0 if np.isnan(ang) else round(ang, 3)
            # 角度特征键值对字典添加拟合角度周期key和对应的拟合角度值
            deg_dict['{}deg_ang{}'.format(
                self.feature_prefix(buy_feature=buy_feature), dk)] = ang
        return deg_dict



# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd
import itertools


class GeneticFactors(object):

    def __init__(self, del_prob, add_prob, change_prob, generations, group_num,
                 sub_num, objective):
        self._del_prob = del_prob  # 删除原始特征概率
        self._add_prob = add_prob  # 添加新特征概率
        self._change_prob = change_prob
        self._generations = generations
        self._sub_num = sub_num
        self._group_num = group_num
        self._objective = objective

    def reset_params(self, del_prob, add_prob, change_prob, generations):
        self._del_prob = del_prob  # 删除原始特征概率
        self._add_prob = add_prob  # 添加新特征概率
        self._change_prob = change_prob
        self._generations = generations

    # 初代种群生成
    def ga_generate_ori(self, ori_field, add_field):
        ## 选取原始特征过程
        ori_list = (np.random.uniform(0, 1, (1, len(ori_field))) <
                    self._del_prob).tolist()[0]
        add_list = (np.random.uniform(0, 1, (1, len(add_field))) <
                    self._add_prob).tolist()[0]
        new_list = ori_list + add_list
        return np.array(new_list)

    # 交叉变异过程
    def ga_cross_next_group(self, ori_group, dict_score='ori'):
        new_dict = ori_group.copy()
        if dict_score == 'ori':
            score = 1.0 / len(ori_group)
            dict_score = {k: score for k in ori_group.keys()}
        g, p = np.array([[k, v] for k, v in dict_score.items()]).T
        flag = max(ori_group.keys())
        ## 按照种群分数进行选择交配
        ## 选择交配种群
        cross_group = np.random.choice(g,
                                       size=int(len(g) / 2),
                                       p=p,
                                       replace=False)
        for (fa, mo) in itertools.combinations(cross_group, 2):
            flag += 1
            fa_code, mo_code = ori_group[fa], ori_group[mo]
            ## 随机选择切分点
            cut_point = np.random.randint(1, len(fa_code) - 1)
            ## 切分基因
            fa_code0, fa_code1 = fa_code[:cut_point], fa_code[cut_point:]
            mo_code0, mo_code1 = mo_code[:cut_point], mo_code[cut_point:]
            ## 基因交换
            new1 = np.hstack([fa_code0, mo_code1])
            ## 变异过程
            prob = np.random.uniform(0, 1)
            if prob < self._change_prob:
                ## 随机挑一个基因点
                change_point = np.random.randint(0, len(new1))
                ## 改变该点的值
                new1[change_point] = not new1[change_point]
            new_dict[flag] = new1
        return new_dict

    ## 种群个体能力评价
    def ga_evalue_group(self, group, total_data, evalue_col):
        score_dict = {}
        for g, code in group.items():
            cols = evalue_col[code]
            score = self._objective(total_data.copy(), cols)
            score_dict[g] = score
        return score_dict

    def ga_kill_group(self, total_data, ori_group, dict_score, evalue_cols):
        ## 二代目
        sub_group = self.ga_cross_next_group(ori_group, dict_score=dict_score)
        ## 评价
        score_dict = self.ga_evalue_group(sub_group, total_data, evalue_cols)
        score_se = pd.Series(score_dict)
        fact_sco = score_se.sort_values(ascending=False)[:self._sub_num]
        score_se = score_se.sort_values(ascending=False)[:self._sub_num] / (
            score_se.sort_values(ascending=False)[:self._sub_num].sum())
        liv_group = {i: sub_group[i] for i in score_dict.keys()}
        return liv_group, dict(score_se), dict(fact_sco)

    def genetic_run(self,
                    total_data,
                    factors_filed=None,
                    strong_field=None,
                    weak_field=None):

        if factors_filed is not None:
            point = int(np.random.uniform(0, len(factors_filed)) / 2)
            ori_field = factors_filed[:point]
            add_field = factors_filed[point:]
        elif strong_field is not None and weak_field is not None:
            ori_field = strong_field
            add_field = weak_field
        evalue_cols = np.array(ori_field + add_field)
        #初始种群
        ori_group = {
            i: self.ga_generate_ori(ori_field, add_field)
            for i in range(self._group_num)
        }
        for i in range(self._generations):
            if i == 0:
                sub, sco, fact_sco = self.ga_kill_group(
                    total_data, ori_group, 'ori', evalue_cols)
            else:
                sub, sco, fact_sco = self.ga_kill_group(
                    total_data, sub, sco, evalue_cols)
            now_sco = sco[(list(sco.keys())[0])]
            now_fact_sco = fact_sco[(list(fact_sco.keys())[0])]
            content_str = "generations:{0},i:{1},sco:{2},fact_sco:{3}".format(
                self._generations, i, now_sco, now_fact_sco)
            print(content_str)
        best_code = pd.Series(sco).sort_values()[-1:].index[0]
        best_field = list(evalue_cols[sub[best_code]])
        return best_field, best_code, sco[best_code]



# -*- coding: utf-8 -*-
from sklearn.covariance import LedoitWolf,ShrunkCovariance,OAS
from sklearn.model_selection import GridSearchCV
import numpy as np


def unshrunk_cov(ret_tb, window=20):
    cov = np.mat(np.cov(ret_tb.tail(window).T.values).astype(float)) * 250
    return cov

def lwshrunk_cov(ret_tb, window=20):
    lw = LedoitWolf()
    cov = lw.fit(ret_tb.tail(window).values).covariance_ * 250
    return cov

def shrunk_cov(ret_tb, window=20):
    cov = ShrunkCovariance().fit(ret_tb.tail(window).values).covariance_ * 250
    return cov

def gridsearch_cov(ret_tb, window=20):
    tuned_parameters = [{'shrinkage': np.logspace(-1.0, 0, 30)}]
    cov = GridSearchCV(ShrunkCovariance(), tuned_parameters).fit(
        ret_tb.tail(window).values).best_estimator_.covariance_ * 250
    return cov

def oas_cov(ret_tb, window=20):
    return OAS().fit(ret_tb.tail(window).values).covariance_ * 250